"""Schema for various output tables produced by hpmcm"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from .cluster import ShearClusterData
from .object import ShearObjectData
from .table import TableColumnInfo, TableInterface

if TYPE_CHECKING:
    from .cell import CellData


SHEAR_NAMES = ["ns", "2p", "2m", "1p", "1m"]


class ObjectAssocTable(TableInterface):
    """Interface of table with associations between objects and sources"""

    _schema = TableInterface._schema.copy()
    _schema.update(
        object_id=TableColumnInfo(int, "Unique Object ID"),
        cluster_id=TableColumnInfo(int, "Parent Cluster Unique ID"),
        source_id=TableColumnInfo(int, "Source id in input catalog"),
        source_idx=TableColumnInfo(int, "Source index in input catalog"),
        catalog_id=TableColumnInfo(int, "Associated catalog ID"),
        distance=TableColumnInfo(float, "Distance from sources to object centroid"),
        cell_idx=TableColumnInfo(int, "Index of associated cell"),
    )

    @staticmethod
    def buildFromCellData(cell_data: CellData) -> ObjectAssocTable:
        """Create object association table

        Parameters
        ----------
        cell_data:
            Cell we are making table for

        Returns
        -------
        Object Association table
        """
        cluster_ids = []
        object_ids = []
        source_ids = []
        source_idxs = []
        cat_idxs = []
        distances_list: list[np.ndarray] = []

        for obj in cell_data.object_dict.values():
            cluster_ids.append(
                np.full((obj.n_src), obj.parent_cluster.i_cluster, dtype=int)
            )
            object_ids.append(np.full((obj.n_src), obj.object_id, dtype=int))
            source_ids.append(obj.sourceIds())
            source_idxs.append(obj.sourceIdxs())
            cat_idxs.append(obj.catalog_id)
            assert obj.dist_2.size
            distances_list.append(obj.dist_2)
        if not distances_list:
            return ObjectAssocTable(
                object_id=np.array([], int),
                cluster_id=np.array([], int),
                source_id=np.array([], int),
                source_idx=np.array([], int),
                catalog_id=np.array([], int),
                distance=np.array([], float),
                cell_idx=np.array([], int),
            )
        distances = np.hstack(distances_list)
        distances = cell_data.matcher.pixToArcsec() * np.sqrt(distances)
        return ObjectAssocTable(
            object_id=np.hstack(object_ids),
            cluster_id=np.hstack(cluster_ids),
            source_id=np.hstack(source_ids),
            source_idx=np.hstack(source_idxs),
            catalog_id=np.hstack(cat_idxs),
            distance=distances,
            cell_idx=np.repeat(cell_data.idx, len(distances)).astype(int),
        )


class ObjectStatsTable(TableInterface):
    """Interface of table of object statistics"""

    _schema = TableInterface._schema.copy()
    _schema.update(
        object_id=TableColumnInfo(int, "Unique Object ID"),
        cluster_id=TableColumnInfo(int, "Parent Cluster Unique ID"),
        n_unique=TableColumnInfo(int, "Number of unique catalogs represented"),
        n_src=TableColumnInfo(int, "Number of sources"),
        dist_rms=TableColumnInfo(
            float, "RMS of distance from sources to object centroid"
        ),
        ra=TableColumnInfo(float, "RA of object centroid"),
        dec=TableColumnInfo(float, "DEC of object centroid"),
        x_cent=TableColumnInfo(float, "X-value of cluster centroid in WCS pixels"),
        y_cent=TableColumnInfo(float, "Y-value of cluster centroid in WCS pixels"),
        snr=TableColumnInfo(float, "Mean signal-to-noise ratio"),
        snr_rms=TableColumnInfo(float, "RMS signal-to-noise ratio"),
        cell_idx=TableColumnInfo(int, "Index of associated cell"),
        has_ref_cat=TableColumnInfo(bool, "Has source from the reference catalog"),
    )

    @staticmethod
    def buildFromCellData(cell_data: CellData) -> ObjectStatsTable:
        """Create object stats table

        Parameters
        ----------
        cell_data:
            Cell we are making table for

        Returns
        -------
        Object stats table
        """
        n_obj = cell_data.n_objects
        cluster_ids = np.zeros((n_obj), dtype=int)
        object_ids = np.zeros((n_obj), dtype=int)
        n_srcs = np.zeros((n_obj), dtype=int)
        n_uniques = np.zeros((n_obj), dtype=int)
        dist_rms = np.zeros((n_obj), dtype=float)
        x_cents = np.zeros((n_obj), dtype=float)
        y_cents = np.zeros((n_obj), dtype=float)
        snrs = np.zeros((n_obj), dtype=float)
        snr_rms = np.zeros((n_obj), dtype=float)
        has_ref_cat = np.zeros((n_obj), dtype=bool)

        for idx, obj in enumerate(cell_data.object_dict.values()):
            cluster_ids[idx] = obj.parent_cluster.i_cluster
            object_ids[idx] = obj.object_id
            n_srcs[idx] = obj.n_src
            n_uniques[idx] = obj.n_unique
            dist_rms[idx] = obj.rms_dist
            x_cents[idx] = obj.x_cent
            y_cents[idx] = obj.y_cent
            assert obj.data is not None
            sum_snr = obj.data.snr.sum()
            x_cents[idx] = np.sum(obj.data.snr * obj.data.x_cell) / sum_snr
            y_cents[idx] = np.sum(obj.data.snr * obj.data.y_cell) / sum_snr
            snrs[idx] = obj.snr_mean
            snr_rms[idx] = obj.snr_rms
            has_ref_cat[idx] = obj.hasRefCatalog()

        ra, dec = cell_data.getRaDec(x_cents, y_cents)
        dist_rms *= cell_data.matcher.pixToArcsec()

        return ObjectStatsTable(
            cluster_id=cluster_ids,
            object_id=object_ids,
            n_unique=n_uniques,
            n_src=n_srcs,
            dist_rms=dist_rms,
            ra=ra,
            dec=dec,
            x_cent=x_cents,
            y_cent=y_cents,
            snr=snrs,
            snr_rms=snr_rms,
            cell_idx=np.repeat(cell_data.idx, len(dist_rms)).astype(int),
            has_ref_cat=has_ref_cat,
        )


class ClusterAssocTable(TableInterface):
    """Interface of table with associations between clusters and sources"""

    _schema = TableInterface._schema.copy()
    _schema.update(
        cluster_id=TableColumnInfo(int, "Unique cluster ID"),
        source_id=TableColumnInfo(int, "Source id in input catalog"),
        source_idx=TableColumnInfo(int, "Source index in input catalog"),
        catalog_id=TableColumnInfo(int, "Associated catalog ID"),
        distance=TableColumnInfo(float, "Distance from sources to cluster centroid"),
        cell_idx=TableColumnInfo(int, "Index of associated cell"),
    )

    @staticmethod
    def buildFromCellData(cell_data: CellData) -> ClusterAssocTable:
        """Create object association table

        Parameters
        ----------
        cell_data:
            Cell we are making table for

        Returns
        -------
        Cluster Association table
        """
        cluster_ids = []
        source_ids = []
        source_idxs = []
        cat_idxs = []
        distances_list: list[np.ndarray] = []
        for cluster in cell_data.cluster_dict.values():
            cluster_ids.append(np.full((cluster.n_src), cluster.i_cluster, dtype=int))
            source_ids.append(cluster.src_id)
            source_idxs.append(cluster.src_idx)
            cat_idxs.append(cluster.catalog_id)
            assert cluster.dist_2.size
            distances_list.append(cluster.dist_2)
        if not distances_list:
            return ClusterAssocTable(
                distance=np.array([], float),
                source_id=np.array([], int),
                source_idx=np.array([], int),
                catalog_id=np.array([], int),
                cluster_id=np.array([], int),
                cell_idx=np.array([], int),
            )
        distances = np.hstack(distances_list)
        distances = cell_data.matcher.pixToArcsec() * np.sqrt(distances)
        return ClusterAssocTable(
            cluster_id=np.hstack(cluster_ids),
            source_id=np.hstack(source_ids),
            source_idx=np.hstack(source_idxs),
            catalog_id=np.hstack(cat_idxs),
            distance=distances,
            cell_idx=np.repeat(cell_data.idx, len(distances)).astype(int),
        )


class ClusterStatsTable(TableInterface):
    """Interface of table of cluster statistics"""

    _schema = TableInterface._schema.copy()
    _schema.update(
        cluster_id=TableColumnInfo(int, "Parent Cluster Unique ID"),
        n_object=TableColumnInfo(int, "Number of objects in cluster"),
        n_unique=TableColumnInfo(int, "Number of unique catalogs represented"),
        n_src=TableColumnInfo(int, "Number of sources"),
        dist_rms=TableColumnInfo(
            float, "RMS of distance from sources to object centroid"
        ),
        ra=TableColumnInfo(float, "RA of object centroid"),
        dec=TableColumnInfo(float, "DEC of object centroid"),
        x_cent=TableColumnInfo(float, "X-value of cluster centroid in WCS pixels"),
        y_cent=TableColumnInfo(float, "Y-value of cluster centroid in WCS pixels"),
        snr=TableColumnInfo(float, "Mean signal-to-noise ratio"),
        snr_rms=TableColumnInfo(float, "RMS signal-to-noise ratio"),
        cell_idx=TableColumnInfo(int, "Index of associated cell"),
        has_ref_cat=TableColumnInfo(bool, "Has source from reference catalog"),
    )

    @staticmethod
    def buildFromCellData(cell_data: CellData) -> ClusterStatsTable:
        """Create object stats table

        Parameters
        ----------
        cell_data:
            Cell we are making table for

        Returns
        -------
        Object stats table
        """
        n_clust = cell_data.n_clusters
        cluster_ids = np.zeros((n_clust), dtype=int)
        n_srcs = np.zeros((n_clust), dtype=int)
        n_uniques = np.zeros((n_clust), dtype=int)
        n_objects = np.zeros((n_clust), dtype=int)
        n_uniques = np.zeros((n_clust), dtype=int)
        dist_rms = np.zeros((n_clust), dtype=float)
        x_cents = np.zeros((n_clust), dtype=float)
        y_cents = np.zeros((n_clust), dtype=float)
        snrs = np.zeros((n_clust), dtype=float)
        snr_rms = np.zeros((n_clust), dtype=float)
        has_ref_cat = np.zeros((n_clust), dtype=bool)

        for idx, cluster in enumerate(cell_data.cluster_dict.values()):
            cluster_ids[idx] = cluster.i_cluster
            n_srcs[idx] = cluster.n_src
            n_uniques[idx] = cluster.n_unique
            n_objects[idx] = len(cluster.objects)
            n_uniques[idx] = cluster.n_unique
            dist_rms[idx] = cluster.rms_dist
            assert cluster.data is not None
            sum_snr = cluster.data.snr.sum()
            x_cents[idx] = np.sum(cluster.data.snr * cluster.data.x_cell) / sum_snr
            y_cents[idx] = np.sum(cluster.data.snr * cluster.data.y_cell) / sum_snr
            snrs[idx] = cluster.snr_mean
            snr_rms[idx] = cluster.snr_rms
            has_ref_cat[idx] = cluster.hasRefCatalog()

        ra, dec = cell_data.getRaDec(x_cents, y_cents)
        dist_rms *= cell_data.matcher.pixToArcsec()

        return ClusterStatsTable(
            cluster_id=cluster_ids,
            n_src=n_srcs,
            n_object=n_objects,
            n_unique=n_uniques,
            dist_rms=dist_rms,
            ra=ra,
            dec=dec,
            x_cent=x_cents,
            y_cent=y_cents,
            snr=snrs,
            snr_rms=snr_rms,
            cell_idx=np.repeat(cell_data.idx, len(dist_rms)).astype(int),
            has_ref_cat=has_ref_cat,
        )


class ShearTable(TableInterface):
    """Interface of table with shear information"""

    _schema = TableInterface._schema.copy()
    _schema["good"] = TableColumnInfo(bool, "Has unique match")
    for _name in SHEAR_NAMES:
        _schema[f"n_{_name}"] = TableColumnInfo(
            float, f"number of sourcrs from catalog {_name}"
        )
        for _i in [1, 2]:
            _schema[f"g_{_i}_{_name}"] = TableColumnInfo(
                float, f"g {_i} for catalog {_name}"
            )
    for _i in [1, 2]:
        for _j in [1, 2]:
            _schema[f"delta_g_{_i}_{_j}"] = TableColumnInfo(
                float, f"delta g {_i} for {_j}p - {_j}m"
            )

    @classmethod
    def buildObjectShearStats(cls, cell_data: CellData) -> ShearTable:
        """Create shear stats table for objects in a cell

        Parameters
        ----------
        cell_data:
            Cell we are making table for

        Returns
        -------
        Shear stats table
        """
        n_obj = cell_data.n_objects
        out_dict = ShearTable.emtpyNumpyDict(n_obj)
        for idx, obj in enumerate(cell_data.object_dict.values()):
            assert isinstance(obj, ShearObjectData)
            obj_stats = obj.shearStats()
            for key, val in obj_stats.items():
                out_dict[key][idx] = val
        return ShearTable(**out_dict)

    @classmethod
    def buildClusterShearStats(cls, cell_data: CellData) -> ShearTable:
        """Create shear stats table for clusters in a cell

        Parameters
        ----------
        cell_data:
            Cell we are making table for

        Returns
        -------
        Shear stats table
        """
        n_clusters = cell_data.n_clusters
        out_dict = ShearTable.emtpyNumpyDict(n_clusters)
        for idx, clus in enumerate(cell_data.cluster_dict.values()):
            assert isinstance(clus, ShearClusterData)
            clus_stats = clus.shearStats()
            for key, val in clus_stats.items():
                out_dict[key][idx] = val
        return ShearTable(**out_dict)



from __future__ import annotations

from typing import Any

import numpy as np
import pandas

from . import input_tables, output_tables
from .cell import CellData, ShearCellData
from .match import Match


# These parameters will have to change if the cells change
TRACT_SIZE = np.array([30000, 30000])
PIXEL_SIZE = 0.2 / 3600.0
CELL_INNER_SIZE = 150
CELL_BUFFER = 25


class ShearMatch(Match):
    """Class to do N-way matching for shear calibration.

    Uses pre-assigned pixel locations from cell-based coadd WCS.

    Since the pixel locations and cells are pre-assigned, the only
    configurable parameters this class takes are the attributres listed
    here.

    The pixel_match_scale can we used to allow for matching sources
    that are seperate by more that 1 pixel.

    Expects 5 input catalogs: a reference catalog and 4 counterfactual
    shear catalogs.

    Attributes
    ----------
    pixel_match_scale: int
        Number of pixels to merge in original counts map

    cat_type: str
        Shear catalog type

    deshear: float | None
        Deshearing parameter, -1*applied shear.  None -> deshearing is not done.

    Notes
    -----
    This expectes a list of parquet files with pandas DataFrames
    that contain the following columns.

    +--------------+---------------------------------------------------------------+
    | Column name  | Description                                                   |
    +==============+===============================================================+
    | id           | source ID                                                     |
    +--------------+---------------------------------------------------------------+
    | tract        | Tract being matched                                           |
    +--------------+---------------------------------------------------------------+
    | x_cell_coadd | X-postion in cell-based coadd used for metadetect             |
    +--------------+---------------------------------------------------------------+
    | y_cell_coadd | Y-postion in cell-based coadd used for metadetect             |
    +--------------+---------------------------------------------------------------+
    | snr          | Signal-to-Noise of source, used for filtering and centroiding |
    +--------------+---------------------------------------------------------------+
    | cell_idx_x   | Cell x-index within Tract                                     |
    +--------------+---------------------------------------------------------------+
    | cell_idx_y   | Cell y-index within Tract                                     |
    +--------------+---------------------------------------------------------------+
    | g_1          | Shear g1 component                                            |
    +--------------+---------------------------------------------------------------+
    | g_2          | Shear g1 component                                            |
    +--------------+---------------------------------------------------------------+

    (see :py:class:`hpmcm.input_tables.ShearCoaddSourceTable`)


    These parquet files can be generated from files with the following
    columns using the ShearMatch.splitByTypeAndClean() function.

    +---------------------------------+---------------------------------------+
    | Column name                     | Description                           |
    +=================================+=======================================+
    | id                              | source ID                             |
    +---------------------------------+---------------------------------------+
    | shear_type                      | one of "ns", "1p", "1m", "2p" "2m"    |
    +---------------------------------+---------------------------------------+
    | patch_{x,y}                     | id of the patch within the tract      |
    +---------------------------------+---------------------------------------+
    | cell_{x,y}                      | id of the cell withing the patch      |
    +---------------------------------+---------------------------------------+
    | snr                             | Signal-to-Noise of source             |
    +---------------------------------+---------------------------------------+
    | {cat_type}_band_flux_{band}     | Flux measuremnt in the reference band |
    +---------------------------------+---------------------------------------+
    | {cat_type}_band_flux_err_{band} | Flux error in the reference band      |
    +---------------------------------+---------------------------------------+
    | {cat_type}_g_{i}                | Shear measurements                    |
    +---------------------------------+---------------------------------------+


    Two additional tables are produced beyond the tables produced by
    the base :py:class:`hpmcm.Match` class

    +----------------+---------------------------------------------------+
    | Key            | Class                                             |
    +================+===================================================+
    | _object_shear  | :py:class:`hpmcm.output_tables.ShearTable`        |
    +----------------+---------------------------------------------------+
    | _cluster_shear | :py:class:`hpmcm.output_tables.ShearTable`        |
    +----------------+---------------------------------------------------+
    """

    inputTableClass: type = input_tables.ShearCoaddSourceTable
    extraCols: list[str] = ["ra", "dec", "x_pix", "y_pix", "g_1", "g_2"]

    def __init__(
        self,
        **kwargs: Any,
    ):
        self.pixel_match_scale: int = kwargs.get("pixel_match_scale", 1)
        self.cat_type: str = kwargs.get("catalogType", "wmom")
        self.deshear: float | None = kwargs.get("deshear", None)
        Match.__init__(self, **kwargs)

    @classmethod
    def createShearMatch(
        cls,
        **kwargs: Any,
    ) -> ShearMatch:
        """Helper function to create a `ShearMatch` object

        This will use the use pixel-coordinates read from
        the input shear tables.

        Parameters
        ----------
        kwargs:
            Passed directly to `ShearMatch` constructor.

        Returns
        -------
        Object to create matches for the requested region
        """
        n_pix = TRACT_SIZE
        kw = dict(
            pixel_size=PIXEL_SIZE,
            n_pixels=n_pix,
            cell_size=CELL_INNER_SIZE,
            cell_buffer=CELL_BUFFER,
            cell_max_object=1000,
        )
        return cls(**kw, **kwargs)

    def getCellIndices(
        self,
        df: pandas.DataFrame,
    ) -> np.ndarray:
        """Get the cell index assocatiated to each source"""
        return (self.n_cell[1] * df["cell_idx_x"] + df["cell_idx_y"]).astype(int)

    def _buildCellData(
        self,
        id_offset: int,
        corner: np.ndarray,
        size: np.ndarray,
        idx: int,
    ) -> CellData:
        return ShearCellData(self, id_offset, corner, size, idx, self.cell_buffer)

    def extractShearStats(self) -> list[pandas.DataFrame]:
        """Extract shear stats

        Theis will produce two :py:class:`hpmcm.output_tables.ShearTable`,
        one for the objects, and the other for the clusters.
        """
        cluster_shear_stats_tables = []
        object_shear_stats_tables = []

        for ix in range(int(self.n_cell[0])):
            for iy in range(int(self.n_cell[1])):
                i_cell = self.getCellIdx(ix, iy)
                if i_cell not in self.cell_dict:
                    continue
                cell_data = self.cell_dict[i_cell]
                assert isinstance(cell_data, ShearCellData)
                cluster_shear_stats_tables.append(
                    output_tables.ShearTable.buildClusterShearStats(cell_data).data
                )
                object_shear_stats_tables.append(
                    output_tables.ShearTable.buildObjectShearStats(cell_data).data
                )

        return [
            pandas.concat(cluster_shear_stats_tables),
            pandas.concat(object_shear_stats_tables),
        ]

    def _getPixValues(self, df: pandas.DataFrame) -> tuple[np.ndarray, np.ndarray]:
        x_pix, y_pix = (
            df["x_pix"].values,
            df["y_pix"].values,
        )
        return x_pix, y_pix

    def reduceDataFrame(
        self,
        df: pandas.DataFrame,
    ) -> pandas.DataFrame:
        """Reduce a single input DataFrame

        Notes
        -----

        This applies a trivial cut on signal-to-noise (snr>1).

        This will add these columns to the output dataframes

        +--------------+-------------------------------------+
        | Column       | Description                         |
        +==============+=====================================+
        | id           | Index of object inside catalog      |
        +--------------+-------------------------------------+
        | ra           | Source RA                           |
        +--------------+-------------------------------------+
        | dec          | Source DEC                          |
        +--------------+-------------------------------------+
        | cell_idx_x   | X-index of Cell                     |
        +--------------+-------------------------------------+
        | cell_idx_y   | Y-index of Cell                     |
        +--------------+-------------------------------------+
        | x_cell_coadd | X-coordinate in cell frame          |
        +--------------+-------------------------------------+
        | y_cell_coadd | Y-coordinate in cell frame          |
        +--------------+-------------------------------------+
        | x_pix        | X-coordinate in global WCS frame    |
        +--------------+-------------------------------------+
        | y_pix        | Y-coordinate in global WCS frame    |
        +--------------+-------------------------------------+
        | g_1          | Shear g_1 component estimate        |
        +--------------+-------------------------------------+
        | g_2          | Shear g_2 component estimate        |
        +--------------+-------------------------------------+
        | snr          | Signal-to-noise ratio               |
        +--------------+-------------------------------------+

        """
        df_clean = df[(df.snr > 1)]
        df_red = df_clean.copy(deep=True)

        return df_red[
            [
                "id",
                "ra",
                "dec",
                "x_pix",
                "y_pix",
                "x_cell_coadd",
                "y_cell_coadd",
                "snr",
                "g_1",
                "g_2",
                "cell_idx_x",
                "cell_idx_y",
            ]
        ]



from __future__ import annotations

from typing import Any

import numpy as np
import pandas
import pyarrow.parquet as pq


class TableColumnInfo:
    """Helper class to manage a column in a table

    This provides a mechanism to document the
    column in the class docstring, and
    to validate input data
    """

    def __init__(self, dtype: type, msg: str):
        self.dtype = dtype
        self.msg = msg

    def __repr__(self) -> str:
        # return f"{self.dtype:10}\n    {self.msg}"
        return f"{self.dtype.__name__:8} | {self.msg:50}"

    def validate(self, val: np.ndarray) -> None:
        """Validate data used to fill a column is of the correct type"""
        assert isinstance(val, np.ndarray)
        assert val.dtype == self.dtype


class TableInterface:
    """Table Schema"""

    _schema: dict[str, TableColumnInfo] = dict()

    def __init__(self, df: pandas.DataFrame | None = None, **kwargs: Any):
        if df is None:
            self._data = self.toPandas(**kwargs)
        else:
            self._data = df

    @classmethod
    def _describeSchema(cls) -> str:
        """Describe the columns in this table"""
        s = []
        for name, val in cls._schema.items():
            assert isinstance(val, TableColumnInfo)
            s.append(f"| {name:15} | {val} |")
        return "\n+-----------------+----------+----------------------------------------------------+\n".join(
            s
        )

    def __init_subclass__(cls, **kwargs: Any) -> None:
        config_text = cls._describeSchema()
        if cls.__doc__ is None:  # pragma: no cover
            cls.__doc__ = f"\nNotes\n-----\n{cls.__name__} schema\n\n"
            cls.__doc__ += "+-----------------+----------+----------------------------------------------------+\n"
            cls.__doc__ += "| Column          | Type     | Description                                        |\n"
            cls.__doc__ += "+=================+==========+====================================================+\n"
            cls.__doc__ += config_text
            cls.__doc__ += "\n+-----------------+----------+----------------------------------------------------+\n"
        else:
            # strip any existing configuration text from parent classes that is at the end of the doctring
            cls.__doc__ = cls.__doc__.split("Notes")[0]
            cls.__doc__ += f"\nNotes\n-----\n{cls.__name__} schema\n\n"
            cls.__doc__ += "+-----------------+----------+----------------------------------------------------+\n"
            cls.__doc__ += "| Column          | Type     | Description                                        |\n"
            cls.__doc__ += "+=================+==========+====================================================+\n"
            cls.__doc__ += config_text
            cls.__doc__ += "\n+-----------------+----------+----------------------------------------------------+\n"

    @property
    def data(self) -> pandas.DataFrame:
        """Return the underlying data"""
        return self._data

    @classmethod
    def validate(cls, **kwargs: Any) -> None:
        """Validate that data match the schema

        Parameters
        ----------
        kwargs:
            The input data

        Raises
        ------
        ValueError:
            The number of columns don't match the schema

        KeyError:
            An input column is not in the schema
        """
        table_size: int = -1
        if len(kwargs) != len(cls._schema):  # pragma: no cover
            raise ValueError(f"{len(kwargs)} != {len(cls._schema)}")
        for key, val in kwargs.items():
            if key not in cls._schema:  # pragma: no cover
                raise KeyError(f"{key} not in {list(cls._schema.keys())}")
            col_info = cls._schema[key]
            col_info.validate(val)
            if table_size < 0:
                table_size = val.size
            else:
                assert val.size == table_size

    @classmethod
    def read(cls, file_path: str, extra_cols: list[str]) -> pandas.DataFrame:
        """Read a dataframe from a file"""
        read_list = list(cls._schema.keys())
        read_list += extra_cols
        parq = pq.read_pandas(file_path, columns=read_list)
        df = parq.to_pandas()
        return df

    @classmethod
    def toPandas(cls, **kwargs: Any) -> pandas.DataFrame:
        """Convert data to a pandas DataFrame

        Parameters
        ----------
        kwargs:
            The input data

        """
        cls.validate(**kwargs)
        return pandas.DataFrame(kwargs)

    @classmethod
    def emtpyNumpyDict(cls, n: int) -> dict[str, np.ndarray]:
        """Create a dict of empty numpy arrays

        Parameter
        ---------
        n:
            Length of the arrays
        """
        return {key: np.zeros((n), dtype=val.dtype) for key, val in cls._schema.items()}



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.figure import Figure, SubFigure

from .cell import CellData
from .cluster import ClusterData
from .match import Match


def showShearObjs(matcher: Match, i_k: tuple[int, int]) -> Figure | SubFigure:
    """Draw the objects in a cluster

    Parameters
    ----------
    matcher:
        Match object

    i_k:
        Indices of the particular cluster

    Returns
    -------
    Figure showing the cluster in question
    """
    cell_data = matcher.cell_dict[i_k[0]]
    cluster = cell_data.cluster_dict[i_k[1]]
    extent = cluster.footprint.extent()
    cluster.extract(cell_data)
    x_offset = cell_data.min_pix[0] + 25
    y_offset = cell_data.min_pix[1] + 25
    assert cluster.data is not None
    x_off = cluster.data.x_pix - x_offset
    y_off = cluster.data.y_pix - y_offset
    catalog_ids = cluster.catalog_id
    image = cluster.footprint.cutout
    img = plt.imshow(image, origin="lower", extent=extent)
    colors = ["red", "blue", "green", "cyan", "orange"]
    markers = [".", "<", ">", "v", "^"]
    for i_obj, obj in enumerate(cluster.objects):
        for x_, y_, i_ in zip(x_off[obj.mask], y_off[obj.mask], catalog_ids[obj.mask]):
            img.axes.scatter(x_, y_, c=colors[i_obj % 5], marker=markers[i_ % 5])
    _cb = plt.colorbar(label="Objects per pixel")
    assert img.axes.figure is not None
    return img.axes.figure


def showShearObj(matcher: Match, i_k: tuple[int, int]) -> Figure | SubFigure:
    """Draw a single object

    Parameters
    ----------
    matcher:
        Match object

    i_k:
        Indices of the particular object

    Returns
    -------
    Figure showing the object in question
    """

    cell_data = matcher.cell_dict[i_k[0]]
    the_obj = cell_data.object_dict[i_k[1]]
    cluster = the_obj.parent_cluster
    extent = cluster.footprint.extent()
    cluster.extract(cell_data)
    x_offset = cell_data.min_pix[0] + 25
    y_offset = cell_data.min_pix[1] + 25
    assert cluster.data is not None
    x_off = cluster.data.x_pix - x_offset
    y_off = cluster.data.y_pix - y_offset
    catalog_ids = cluster.catalog_id
    img = plt.imshow(cluster.footprint.cutout, origin="lower", extent=extent)
    markers = [".", "<", ">", "v", "^"]
    for _i_obj, obj in enumerate(cluster.objects):
        if obj.object_id == the_obj.object_id:
            color = "red"
        else:  # pragma: no cover
            color = "blue"
        for x_, y_, i_ in zip(x_off[obj.mask], y_off[obj.mask], catalog_ids[obj.mask]):
            img.axes.scatter(x_, y_, c=color, marker=markers[i_ % 5])
    _cb = plt.colorbar(label="Objects per pixel")
    assert img.axes.figure is not None
    return img.axes.figure


def showCluster(
    image: np.ndarray,
    cluster: ClusterData,
    cell_data: CellData,
    mask: np.ndarray | None = None,
) -> Figure | SubFigure:
    """Draw a cluster

    Parameters
    ----------
    image:
        Counts map used to make clusters

    cluster:
        Cluster being draw

    cell_data:
        Parent Cell for the cluster

    mask:
        Mask showing which sources are in the cluster

    Returns
    -------
    Figure showing the cluster in question
    """
    extent = cluster.footprint.extent()
    cluster.extract(cell_data)
    x_offset = cell_data.min_pix[0] + 25
    y_offset = cell_data.min_pix[1] + 25
    x_off = cluster.x_cluster
    y_off = cluster.y_cluster
    if mask is not None:  # pragma: no cover
        x_off = x_off[mask]
        y_off = y_off[mask]
    x_c = cluster.x_cent - x_offset
    y_c = cluster.y_cent - y_offset

    img = plt.imshow(
        image[cluster.footprint.slice_x][cluster.footprint.slice_y],
        origin="lower",
        extent=extent,
        cmap="grey",
    )
    _cb = plt.colorbar(label="Objects per pixel")
    try:
        assert cluster.data is not None
        x_off_u = x_off - cluster.data.dxShear
        y_off_u = y_off - cluster.data.dyShear
        img.axes.scatter(x_off_u, y_off_u, marker="x")
    except Exception:  # pragma: no cover
        pass
    img.axes.scatter(x_off, y_off)
    img.axes.scatter(x_c, y_c, marker="+", c="green")
    img.axes.set_xlabel("x [pixels]")
    img.axes.set_ylabel("y [pixels]")
    assert img.axes.figure is not None
    return img.axes.figure


def showObjects(
    image: np.ndarray,
    cluster: ClusterData,
    cell_data: CellData,
) -> Figure | SubFigure:
    """Draw a cluster, showing the objects

    Parameters
    ----------
    image:
        Counts map used to make clusters

    cluster:
        Cluster being draw

    cell_data:
        Parent Cell for the cluster

    mask:
        Mask showing which sources are in the cluster

    Returns
    -------
    Figure showing the cluster in question
    """
    extent = cluster.footprint.extent()
    cluster.extract(cell_data)
    x_offset = cell_data.min_pix[0]
    y_offset = cell_data.min_pix[1]
    x_off = cluster.x_cluster
    y_off = cluster.y_cluster
    img = plt.imshow(
        image[cluster.footprint.slice_x][cluster.footprint.slice_y],
        origin="lower",
        extent=extent,
    )
    _cb = plt.colorbar(label="Objects per pixel")
    colors = ["red", "blue", "green", "cyan", "orange"]
    for i_obj, obj in enumerate(cluster.objects):
        x_c = obj.x_cent - x_offset
        y_c = obj.y_cent - y_offset
        img.axes.scatter(
            x_off[obj.mask],
            y_off[obj.mask],
            c=colors[i_obj % 5],
            s=1 + np.ceil(i_obj / 5),
        )
        print(1 + np.ceil(i_obj / 5))
        img.axes.scatter(x_c, y_c, marker="+", c=colors[i_obj % 6])
    assert img.axes.figure is not None
    return img.axes.figure


def showObjectsV2(
    image: np.ndarray,
    cluster: ClusterData,
    cell_data: CellData,
) -> Figure | SubFigure:
    """Draw a cluster, showing the objects

    Parameters
    ----------
    image:
        Counts map used to make clusters

    cluster:
        Cluster being draw

    cell_data:
        Parent Cell for the cluster

    Returns
    -------
    Figure showing the cluster in question
    """
    extent = cluster.footprint.extent()
    cluster.extract(cell_data)
    x_offset = cell_data.min_pix[0]
    y_offset = cell_data.min_pix[1]
    x_off = cluster.x_pix - x_offset
    y_off = cluster.y_pix - y_offset
    img = plt.imshow(
        image[cluster.footprint.slice_x][cluster.footprint.slice_y],
        origin="lower",
        extent=extent,
    )
    _cb = plt.colorbar()
    colors = ["red", "blue", "green", "cyan", "orange"]
    for x_off_, y_off_, i_cat_ in zip(x_off, y_off, cluster.catalog_id):
        if i_cat_ % 5 == 0 and i_cat_ != 20:
            continue
        img.axes.scatter(
            x_off_, y_off_, c=colors[i_cat_ % 5], s=20 - 3 * np.ceil(i_cat_ / 5)
        )
    assert img.axes.figure is not None
    return img.axes.figure



"""Expected schema for various input tables used by hpmcm"""

from __future__ import annotations

from .table import TableColumnInfo, TableInterface


class SourceTable(TableInterface):
    """Basic input table Using RA and DEC"""

    _schema = TableInterface._schema.copy()
    _schema.update(
        id=TableColumnInfo(int, "Unique ID for source"),
        ra=TableColumnInfo(float, "RA of source"),
        dec=TableColumnInfo(float, "DEC of source"),
        snr=TableColumnInfo(float, "Signal-to-noise of source"),
    )


class CoaddSourceTable(TableInterface):
    """Basic input table using cell-positions"""

    _schema = TableInterface._schema.copy()
    _schema.update(
        id=TableColumnInfo(int, "Unique ID for source"),
        tract=TableColumnInfo(int, "Tract"),
        x_cell_coadd=TableColumnInfo(
            float, "X-postion in cell-based coadd used for metadetect"
        ),
        y_cell_coadd=TableColumnInfo(
            float, "Y-postion in cell-based coadd used for metadetect"
        ),
        snr=TableColumnInfo(float, "Signal-to-noise of source"),
        cell_idx_x=TableColumnInfo(int, "Cell x-index within Tract"),
        cell_idx_y=TableColumnInfo(int, "Cell y-index within Tract"),
    )


class ShearCoaddSourceTable(CoaddSourceTable):
    """Shear calibration input table using cell-positions"""

    _schema = CoaddSourceTable._schema.copy()
    _schema.update(
        g_1=TableColumnInfo(float, "Shear g1 component"),
        g_2=TableColumnInfo(float, "Shear g2 component"),
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas

from . import shear_utils
from .footprint import Footprint
from .object import ObjectData

if TYPE_CHECKING:
    from .cell import CellData


class ClusterData:
    """Class that defines clusters

    A cluster is a set of sources within a footprint of
    adjacent pixels.

    Attributes
    ----------
    i_cluster : int
        Cluster ID

    footprint: Footprint
        Footprint of this cluster in the CellData counts map

    orig_cluster : int
        Id of the original cluster this cluster was made from

    sources: np.ndarray
        Data about the sources in this cluster

    n_src : int
        Number of sources in this cluster

    n_unique : int
        Number of catalogs contributing sources to this cluster

    objects: list[ObjectData]
        Data about the objects in this cluster

    x_cent : float
        X-pixel value of cluster centroid (in WCS used to do matching)

    y_cent : float
        Y-pixel value of cluster centroid (in WCS used to do matching)

    rms_dist: float
        RMS distance of sources to the centroid

    snr_mean: float
        mean signal-to-noise

    snr_rms: float
        RMS of signal-to-noise

    dist_2: np.ndarray
        Distances of sources to the centroid
    """

    def __init__(
        self,
        i_cluster: int,
        footprint: Footprint,
        sources: np.ndarray,
        orig_cluster: int | None = None,
    ):
        """Build from a Footprint and data about the
        sources in that Footprint"""
        self.i_cluster: int = i_cluster
        self.footprint: Footprint = footprint
        self.orig_cluster: int = i_cluster
        if orig_cluster is not None:
            self.orig_cluster = orig_cluster
        self.sources: np.ndarray = sources

        self.n_src: int = self.sources[0].size
        self.n_unique: int = len(np.unique(self.sources[0]))
        self.objects: list[ObjectData] = []
        self.data: pandas.DataFrame | None = None

        self.x_cent: float = np.nan
        self.y_cent: float = np.nan
        self.rms_dist: float = np.nan
        self.snr_mean: float = np.nan
        self.snr_rms: float = np.nan
        self.dist_2: np.ndarray = np.array([])
        self.pixel_match_scale: int = 1

    def extract(self, cell_data: CellData) -> None:
        """Extract the x_pix, y_pix and snr data from
        the sources in this cluster
        """
        x_offset = self.footprint.slice_x.start * self.pixel_match_scale
        y_offset = self.footprint.slice_y.start * self.pixel_match_scale

        series_list = []

        for _i, (i_cat_, src_idx_) in enumerate(zip(self.sources[0], self.sources[2])):
            series_list.append(cell_data.data[i_cat_].iloc[src_idx_])

        self.data = pandas.DataFrame(series_list)
        self.data["i_cat"] = self.sources[0]
        self.data["src_id"] = self.sources[1]
        self.data["src_idx"] = self.sources[2]
        self.data["x_cluster"] = self.data.x_cell - x_offset
        self.data["y_cluster"] = self.data.y_cell - y_offset

    @property
    def catalog_id(self) -> np.ndarray:
        """Return the catalog indices for all the sources in the cluster"""
        assert self.data is not None
        return self.data.i_cat

    @property
    def src_id(self) -> np.ndarray:
        """Return the ids for all the sources in the cluster"""
        assert self.data is not None
        return self.data.src_id

    @property
    def src_idx(self) -> np.ndarray:
        """Return the indices for all the sources in the cluster"""
        assert self.data is not None
        return self.data.src_idx

    def hasRefCatalog(self, ref_cat_id: int = 0) -> bool:
        """Is there a source from the reference catalog"""
        return ref_cat_id in self.sources[0]

    @property
    def x_cluster(self) -> np.ndarray:
        """Return the x-positions of the soures w.r.t. the footprint"""
        assert self.data is not None
        return self.data.x_cluster

    @property
    def y_cluster(self) -> np.ndarray:
        """Return the y-positions of the soures w.r.t. the footprint"""
        assert self.data is not None
        return self.data.y_cluster

    @property
    def x_pix(self) -> np.ndarray:
        """Return the x-positions of the soures w.r.t. the WCS"""
        assert self.data is not None
        return self.data.x_pix

    @property
    def y_pix(self) -> np.ndarray:
        """Return the x-positions of the soures w.r.t. the WCS"""
        assert self.data is not None
        return self.data.y_pix

    def addObject(
        self, cell_data: CellData, mask: np.ndarray | None = None
    ) -> ObjectData:
        """Add a new object to this cluster"""
        new_object = cell_data.addObject(self, mask)
        self.objects.append(new_object)
        return new_object


class ShearClusterData(ClusterData):
    """Subclass of ClusterData that can compute shear statisitics

    Attributes
    ----------
    pixel_match_scale: int
        Number of pixel merged in the original counts map
    """

    def __init__(
        self,
        i_cluster: int,
        footprint: Footprint,
        sources: np.ndarray,
        orig_cluster: int | None = None,
        pixel_match_scale: int = 1,
    ):
        ClusterData.__init__(self, i_cluster, footprint, sources, orig_cluster)
        self.pixel_match_scale = pixel_match_scale

    def shearStats(self) -> dict:
        """Return the shear statistics"""
        assert self.data is not None
        return shear_utils.shearStats(self.data)



from __future__ import annotations

import numpy as np
import pandas
import tables_io

from .footprint import FootprintSet


def findClusterIdsFromArrays(
    x_locals: np.ndarray,
    y_locals: np.ndarray,
    cluster_key: np.ndarray,
) -> np.ndarray:
    """Associate sources to clusters using `clusterkey`
    which is a map where any pixel associated to a cluster
    has the cluster index as its value

    Parameters
    ----------
    x_locals:
        Local pixel x-positions

    y_locals:
        Local pixel y-positions

    cluster_key:
        2D-map of cluster Ids by pixel position

    Returns
    -------
    Ids of associated clusters
    """
    return np.array(
        [
            cluster_key[x_local_, y_local_]
            for x_local_, y_local_ in zip(x_locals, y_locals)
        ]
    ).astype(np.int32)


def findClusterIds(
    df: pandas.DataFrame,
    cluster_key: np.ndarray,
    pixel_match_scale: int = 1,
) -> np.ndarray:
    """Associate sources to clusters using `clusterkey`
    which is a map where any pixel associated to a cluster
    has the cluster index as its value

    Parameters
    ----------
    df:
        DataFrame with local pixel positions (x_cell, y_cell)

    cluster_key:
        2D-map of cluster Ids by pixel position

    pixel_match_scale:
        Scale-factor to use in making cluster map

    Returns
    -------
    Ids of associated clusters
    """
    return findClusterIdsFromArrays(
        np.floor(df["x_cell"] / pixel_match_scale).astype(int),
        np.floor(df["y_cell"] / pixel_match_scale).astype(int),
        cluster_key,
    )


def fillCountsMapFromArrays(
    x_locals: np.ndarray,
    y_locals: np.ndarray,
    n_pix: np.ndarray,
    weights: np.ndarray | None = None,
) -> np.ndarray:
    """Fill a source counts map

    Parameters
    ----------
    x_locals:
        Local pixel x-positions

    y_locals:
        Local pixel y-positions

    n_pix:
        Number of pixels in x,y for counts map

    weights:
        If provided, weights to apply for each entry in counts map

    Returns
    -------
    Counts map of source in cell, projected into n_pix,n_pix grid
    """
    hist = np.histogram2d(
        x_locals,
        y_locals,
        bins=(n_pix[0], n_pix[1]),
        range=((0, n_pix[0]), (0, n_pix[1])),
        weights=weights,
    )
    return hist[0]


def fillCountsMapFromDf(
    df: pandas.DataFrame,
    n_pix: np.ndarray,
    weight_name: str | None = None,
    pixel_match_scale: int = 1,
) -> np.ndarray:
    """Fill a source counts map from a reduced dataframe for one input
    catalog

    Parameters
    ----------
    df:
        DataFrame with local pixel positions (x_cell, y_cell)

    n_pix:
        Number of pixels in x,y for counts map

    weight_name:
        If provided column to use for weights

    pixel_match_scale:
        Scale-factor to use in making cluster map

    Returns
    -------
    Counts map of source in cell, projected into n_pix,n_pix grid
    """
    if weight_name is None:
        weights = None
    else:  # pragma: no cover
        weights = df[weight_name].values
    return fillCountsMapFromArrays(
        df["x_cell"] / pixel_match_scale,
        df["y_cell"] / pixel_match_scale,
        n_pix=np.ceil(n_pix / pixel_match_scale).astype(int),
        weights=weights,
    )


def getFootprints(
    counts_map: np.ndarray,
    buf: int,
    pixel_match_scale: int = 1,
) -> dict:
    """Take a source counts map and do clustering using Footprint detection

    Parameters
    ----------
    counts_map:
        Map of source counts

    buf:
        Number of pixels in cell-edge buffer

    pixel_match_scale:
        Scale-factor used in making cluster map

    Returns
    -------
    Footprint data

    +---------------+------------------+---------------------------------+
    | Key           | Type             | Description                     |
    +===============+==================+=================================+
    | image         | np.ndarray       | Counts map of sources           |
    +---------------+------------------+---------------------------------+
    | footprints    | FootprintSet     | Clustering footprints           |
    +---------------+------------------+---------------------------------+
    | footprint_key | np.ndarray       | Array with cluster associations |
    +---------------+------------------+---------------------------------+

    """
    footprints_orig = FootprintSet.detect(counts_map)
    footprints = footprints_orig.filter(buf, pixel_match_scale)
    footprint_key = footprints.fp_key
    return dict(image=counts_map, footprints=footprints, footprint_key=footprint_key)


def associateSourcesToFootprints(
    data: list[pandas.DataFrame],
    cluster_key: np.ndarray,
    pixel_match_scale: int = 1,
) -> list[np.ndarray]:
    """Loop through data and associate sources to clusters

    Parameters
    ----------
    data:
        Input DataFrames

    cluster_key:
        2D-map of cluster Ids by pixel position

    pixel_match_scale:
        Scale-factor used in making cluster map

    Returns
    -------
    Lists of clusters associated to each source
    output[i][j] will give the id of the cluster associated
    to source j in input catalog i.
    """
    return [findClusterIds(df, cluster_key, pixel_match_scale) for df in data]


def reduceObjectTable(
    basefile: str,
    outfile: str,
    extra_cols: list[str] | None = None,
) -> None:  # pragma: no cover
    """Reduce an object table to just the colums needed for matching

    Parameters
    ----------
    basefile:
        Original file name

    outfile:
        Output file name

    extra_cols:
        Extra columns to copy

    Notes
    -----
    This will produce a DataFrame with at least these columns:

    +-----------------------+---------------------------------------------------------------+
    | Column name           | Description                                                   |
    +=======================+===============================================================+
    | id                    | source ID                                                     |
    +-----------------------+---------------------------------------------------------------+
    | tract                 | Tract source was found in                                     |
    +-----------------------+---------------------------------------------------------------+
    | patch                 | Patch source was found in                                     |
    +-----------------------+---------------------------------------------------------------+
    | ra                    | RA in degrees                                                 |
    +-----------------------+---------------------------------------------------------------+
    | dec                   | DEC in degress                                                |
    +-----------------------+---------------------------------------------------------------+
    | snr                   | Signal-to-Noise of source, used for filtering and centroiding |
    +-----------------------+---------------------------------------------------------------+
    | {band}_gaapPsfFlux    | Flux, for band in u,g,r,i,z,y                                 |
    +-----------------------+---------------------------------------------------------------+
    | {band}_gaapPsfFluxErr | Flux error, for band in u,g,r,i,z,y                           |
    +-----------------------+---------------------------------------------------------------+

    """
    t = tables_io.read(basefile)
    cols = ["tract", "patch", "coord_ra", "coord_dec", "objectId"]
    cols += [f"{band}_gaapPsfFlux" for band in "ugrizy"]
    cols += [f"{band}_gaapPsfFluxErr" for band in "ugrizy"]

    if extra_cols is not None:
        cols += extra_cols

    tout = t[cols].copy(deep=True)

    tout["ra"] = tout["coord_ra"]
    tout["dec"] = tout["coord_dec"]
    tout["snr"] = np.where(
        np.isfinite(tout["r_gaapPsfFlux"]),
        tout["r_gaapPsfFlux"] / tout["i_gaapPsfFluxErr"],
        0,
    )
    tout["id"] = tout["objectId"]

    tout.to_parquet(outfile)



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, Any

import numpy as np

if TYPE_CHECKING:
    from .cluster import ClusterData
    from .match import Match


def clusterStats(cluster_dict: OrderedDict[int, ClusterData]) -> np.ndarray:
    """Helper function to get stats about the clusters

    Parameters
    ----------
    cluster_dict:
        Dict from clusterId to ClusterData object

    Returns
    -------
    Cluster Statistics (nClusters, n_orphan, n_mixed, n_confused)

    Notes
    -----
    Return array contains

    nClusters: Total number of clusters

    n_orphan: Number of single source clusters (i.e., single detections)

    n_mixed: Number of clusters with more than one source from each input catalog

    n_confused: Number of souces with more than four cases of duplication
    """
    n_orphan = 0
    n_mixed = 0
    n_confused = 0
    for val in cluster_dict.values():
        if val.n_src == 1:
            n_orphan += 1
        if val.n_src != val.n_unique:
            n_mixed += 1
            if val.n_src > val.n_unique + 3:  # pragma: no cover
                n_confused += 1
    return np.array([len(cluster_dict), n_orphan, n_mixed, n_confused])


def printSummaryStats(matcher: Match) -> np.ndarray:
    """Helper function to print info about clusters"""
    stats = np.zeros((4), int)
    for key, cell_data in matcher.cell_dict.items():
        cell_stats = clusterStats(cell_data.cluster_dict)
        print(
            f"{key:5}: "
            f"{cell_stats[0]:5} "
            f"{cell_stats[1]:5} "
            f"{cell_stats[2]:5} "
            f"{cell_stats[3]:5}"
        )
        stats += cell_stats
    return stats


def classifyClusters(matcher: Match, **kwargs: Any) -> dict[str, list]:
    """Sort clusters by their properties

    This will return a dict of lists of clusters of various types
    """
    nsrcs = []

    cut1 = []
    cut2 = []

    used = []
    ideal_faint = []
    ideal = []

    faint = []
    edge_mixed = []
    mixed = []
    edge_missing = []
    edge_extra = []

    missing = []
    two_missing = []
    many_missing = []
    extra = []
    caught = []

    cell_edge = kwargs.get("cell_edge", 75)
    edge_cut = kwargs.get("edge_cut", 2)
    snr_cut = kwargs.get("snr_cut", 7.5)

    n_cat = len(matcher.red_data)

    for i_c, cell_data in matcher.cell_dict.items():
        cd = cell_data.cluster_dict

        for key, c in cd.items():
            k = (i_c, key)

            assert c.data is not None

            nsrcs.append(c.n_src)

            if (np.fabs(c.data.x_cell_coadd) > cell_edge).all() or (
                np.fabs(c.data.y_cell_coadd) > cell_edge
            ).all():
                cut1.append(k)
                continue
            if (
                np.fabs(c.data.x_cell_coadd.mean()) > cell_edge
                or np.fabs(c.data.y_cell_coadd.mean()) > cell_edge
            ):
                cut2.append(k)
                continue

            used.append(k)

            edge_case = False
            is_faint = False
            if (np.fabs(c.data.x_cell_coadd) > cell_edge - edge_cut).any() or (
                np.fabs(c.data.y_cell_coadd) > cell_edge - edge_cut
            ).any():
                edge_case = True
            if (c.data.snr < snr_cut).any():
                is_faint = True

            if c.n_src == c.n_unique and c.n_src == n_cat and is_faint:
                ideal_faint.append(k)
            elif c.n_src == c.n_unique and c.n_src == n_cat:
                ideal.append(k)
            elif c.n_src < n_cat and is_faint:
                faint.append(k)
            elif (
                c.n_src == n_cat and c.n_unique != n_cat and edge_case
            ):  # pragma: no cover
                edge_mixed.append(k)
            elif c.n_src == n_cat and c.n_unique != n_cat:  # pragma: no cover
                mixed.append(k)
            elif c.n_src < n_cat and edge_case:  # pragma: no cover
                edge_missing.append(k)
            elif c.n_src > n_cat and edge_case:  # pragma: no cover
                edge_extra.append(k)
            elif c.n_src == n_cat - 1:  # pragma: no cover
                missing.append(k)
            elif c.n_src == n_cat - 2:  # pragma: no cover
                two_missing.append(k)
            elif c.n_src < n_cat - 2:
                many_missing.append(k)
            elif c.n_src > n_cat:
                extra.append(k)
            else:  # pragma: no cover
                caught.append(k)

    return dict(
        nsrcs=nsrcs,
        cut1=cut1,
        cut2=cut2,
        used=used,
        ideal_faint=ideal_faint,
        ideal=ideal,
        faint=faint,
        edge_mixed=edge_mixed,
        mixed=mixed,
        edge_missing=edge_missing,
        edge_extra=edge_extra,
        missing=missing,
        two_missing=two_missing,
        many_missing=many_missing,
        extra=extra,
        caught=caught,
    )


def matchObjectsAgainstRef(matcher: Match, **kwargs: Any) -> dict[str, list]:
    """Match objects against the reference catalog"""
    nsrcs = []
    used = []

    ideal_faint = []
    ideal = []

    faint = []
    not_in_ref = []
    not_in_ref_faint = []
    in_ref = []

    extra = []
    missing = []
    two_missing = []
    many_missing = []
    caught = []

    snr_cut = kwargs.get("snr_cut", 7.5)

    n_cat = len(matcher.red_data)

    for i_c, cell_data in matcher.cell_dict.items():
        od = cell_data.object_dict
        for key, c in od.items():
            k = (i_c, key)

            assert c.data is not None
            nsrcs.append(c.n_src)
            used.append(k)

            is_faint = False
            if (c.data.snr < snr_cut).any():
                is_faint = True

            if (c.catalog_id == 0).any():
                in_ref.append(k)
            else:
                if is_faint:
                    not_in_ref_faint.append(k)
                else:
                    not_in_ref.append(k)
                continue

            if c.n_src == c.n_unique and c.n_src == n_cat and is_faint:
                ideal_faint.append(k)
            elif c.n_src == c.n_unique and c.n_src == n_cat:
                ideal.append(k)
            elif is_faint:
                faint.append(k)
            elif c.n_src == n_cat - 1:
                missing.append(k)
            elif c.n_src == n_cat - 2:  # pragma: no cover
                two_missing.append(k)
            elif c.n_src < n_cat - 2:  # pragma: no cover
                many_missing.append(k)
            elif c.n_src > n_cat:
                extra.append(k)
            else:  # pragma: no cover
                caught.append(k)

    return dict(
        nsrcs=nsrcs,
        used=used,
        ideal_faint=ideal_faint,
        ideal=ideal,
        faint=faint,
        missing=missing,
        in_ref=in_ref,
        not_in_ref=not_in_ref,
        not_in_ref_faint=not_in_ref_faint,
        extra=extra,
        two_missing=two_missing,
        many_missing=many_missing,
        caught=caught,
    )


def printObjectMatchTypes(o_dict: dict) -> None:
    """Print numbers of different types of object matches"""
    print("All            ", len(o_dict["nsrcs"]))
    print("Used           ", len(o_dict["used"]))
    print("  New            ", len(o_dict["not_in_ref"]))
    print("  New (faint)    ", len(o_dict["not_in_ref_faint"]))
    print("In Ref         ", len(o_dict["in_ref"]))
    print("Faint          ", len(o_dict["faint"]))
    print("Good           ", len(o_dict["ideal"]))
    print("  Good (faint)   ", len(o_dict["ideal_faint"]))
    print("Missing        ", len(o_dict["missing"]))
    print("Two Missing    ", len(o_dict["two_missing"]))
    print("All Missing    ", len(o_dict["many_missing"]))
    print("Extra          ", len(o_dict["extra"]))
    print("Caught         ", len(o_dict["caught"]))


def classifyObjects(matcher: Match, **kwargs: Any) -> dict[str, list]:
    """Sort objects by their properties

    This will return a dict of lists of objects
    """

    nsrcs = []

    cut1 = []
    cut2 = []

    used = []
    ideal_faint = []
    ideal = []

    faint = []
    edge_mixed = []
    mixed = []
    edge_missing = []
    edge_extra = []

    orphan = []
    missing = []
    two_missing = []
    many_missing = []
    extra = []
    caught = []

    cell_edge = kwargs.get("cell_edge", 75)
    edge_cut = kwargs.get("edge_cut", 2)
    snr_cut = kwargs.get("snr_cut", 7.5)

    n_cat = len(matcher.red_data)

    for i_c, cell_data in matcher.cell_dict.items():
        od = cell_data.object_dict

        for key, c in od.items():
            k = (i_c, key)

            assert c.data is not None

            nsrcs.append(c.n_src)

            try:
                if (np.fabs(c.data.x_cell_coadd) > cell_edge).all() or (
                    np.fabs(c.data.y_cell_coadd) > cell_edge
                ).all():
                    cut1.append(k)
                    continue
            except Exception:
                pass
            try:
                if (
                    np.fabs(c.data.x_cell_coadd.mean()) > cell_edge
                    or np.fabs(c.data.y_cell_coadd.mean()) > cell_edge
                ):
                    cut2.append(k)
                    continue
            except Exception:
                pass

            used.append(k)

            edge_case = False
            is_faint = False
            try:
                if (np.fabs(c.data.x_cell_coadd) > cell_edge - edge_cut).any() or (
                    np.fabs(c.data.y_cell_coadd) > cell_edge - edge_cut
                ).any():
                    edge_case = True
            except Exception:
                edge_case = False
            if (c.data.snr < snr_cut).any():
                is_faint = True
            if c.n_src == c.n_unique and c.n_src == n_cat and is_faint:
                ideal_faint.append(k)
            elif c.n_src == c.n_unique and c.n_src == n_cat:
                ideal.append(k)
            elif c.n_src < n_cat and is_faint:
                faint.append(k)
            elif (
                c.n_src == n_cat and c.n_unique != n_cat and edge_case
            ):  # pragma: no cover
                edge_mixed.append(k)
            elif c.n_src == n_cat and c.n_unique != n_cat:
                mixed.append(k)
            elif c.n_src < n_cat and edge_case:  # pragma: no cover
                edge_missing.append(k)
            elif (
                c.n_src < n_cat and c.parent_cluster.n_src >= n_cat
            ):  # pragma: no cover
                orphan.append(k)
            elif c.n_src == n_cat - 1:
                missing.append(k)
            elif c.n_src == n_cat - 2:  # pragma: no cover
                two_missing.append(k)
            elif c.n_src < n_cat - 2:
                many_missing.append(k)
            elif c.n_src > n_cat and edge_case:  # pragma: no cover
                edge_extra.append(k)
            elif c.n_src > n_cat:
                extra.append(k)
            else:  # pragma: no cover
                caught.append(k)

    return dict(
        nsrcs=nsrcs,
        cut1=cut1,
        cut2=cut2,
        used=used,
        ideal_faint=ideal_faint,
        ideal=ideal,
        faint=faint,
        edge_mixed=edge_mixed,
        mixed=mixed,
        edge_missing=edge_missing,
        edge_extra=edge_extra,
        orphan=orphan,
        missing=missing,
        two_missing=two_missing,
        many_missing=many_missing,
        extra=extra,
        caught=caught,
    )


def printClusterTypes(cluster_types: dict[str, list]) -> None:
    """Print numbers of different types of clusters"""
    print(
        "All Clusters:                                  ",
        len(cluster_types["nsrcs"]),
    )
    print("cut 1                                          ", len(cluster_types["cut1"]))
    print("cut 2                                          ", len(cluster_types["cut2"]))
    print("Used:                                          ", len(cluster_types["used"]))
    print(
        "good (n source from n catalogs):               ",
        len(cluster_types["ideal"]),
    )
    print(
        "good faint                                     ",
        len(cluster_types["ideal_faint"]),
    )
    print(
        "faint (< n sources, snr < cut):                ",
        len(cluster_types["faint"]),
    )
    print(
        "mixed (n source from < n catalogs):            ",
        len(cluster_types["mixed"]),
    )
    print(
        "edge_mixed (mixed near edge of cell):          ",
        len(cluster_types["edge_mixed"]),
    )
    print(
        "edge_missing (< n sources, near edge of cell): ",
        len(cluster_types["edge_missing"]),
    )
    print(
        "edge_extra (> n sources, near edge of cell):   ",
        len(cluster_types["edge_extra"]),
    )
    print(
        "faint (< n sources, snr < cut):                ",
        len(cluster_types["faint"]),
    )
    print(
        "one missing (n-1 sources, not near edge):      ",
        len(cluster_types["missing"]),
    )
    print(
        "two missing (n-2 sources, not near edge):      ",
        len(cluster_types["two_missing"]),
    )
    print(
        "many missing (< n-2 sources, not near edge):   ",
        len(cluster_types["many_missing"]),
    )
    print(
        "extra (> n sources, not near edge):            ",
        len(cluster_types["extra"]),
    )


def printObjectTypes(object_types: dict[str, list]) -> None:
    """Print numbers of different types of objects"""
    print("All Objects:                                   ", len(object_types["nsrcs"]))
    print("cut 1                                          ", len(object_types["cut1"]))
    print("cut 2                                          ", len(object_types["cut2"]))
    print("Used:                                          ", len(object_types["used"]))
    print("good (n source from n catalogs):               ", len(object_types["ideal"]))
    print(
        "good faint                                     ",
        len(object_types["ideal_faint"]),
    )
    print("faint (< n sources, snr < cut):                ", len(object_types["faint"]))
    print("mixed (n source from < n catalogs):            ", len(object_types["mixed"]))
    print(
        "edge_mixed (mixed near edge of cell):          ",
        len(object_types["edge_mixed"]),
    )
    print(
        "edge_missing (< n sources, near edge of cell): ",
        len(object_types["edge_missing"]),
    )
    print(
        "edge_extra (> n sources, near edge of cell):   ",
        len(object_types["edge_extra"]),
    )
    print("faint (< n sources, snr < cut):                ", len(object_types["faint"]))
    print(
        "orphan (split off from larger cluster          ",
        len(object_types["orphan"]),
    )
    print(
        "one missing (n-1 sources, not near edge):      ",
        len(object_types["missing"]),
    )
    print(
        "two missing (n-2 sources, not near edge):      ",
        len(object_types["two_missing"]),
    )
    print(
        "many missing (< n-2 sources, not near edge):   ",
        len(object_types["many_missing"]),
    )
    print("extra (> n sources, not near edge):            ", len(object_types["extra"]))



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from . import utils
from .cluster import ClusterData
from .object import ObjectData

if TYPE_CHECKING:
    from .cell import CellData


def heirarchicalProcessObject(
    obj_data: ObjectData, cell_data: CellData, pixel_r2_cut: float, recurse: int = 0
) -> None:
    """Recursively process an object and make sub-objects

    Parameters
    ----------
    obj_data:
        Object being processed

    cell_data:
        Associated cell

    pixel_r2_cut:
        Distance cut for associated, in pixels**2

    recurse:
        Current recursion level


    Notes
    -----
    This function will test if the input obj_data is good,
    i.e., that all the sources lie within the pixel_r2_cut,
    and will split obj_data into multiple objects if
    some of the sources lie outside the cut

    The new objects will be added to the parent cluster of
    the original object
    """
    if recurse > cell_data.matcher.max_sub_division:
        return
    obj_data.recurse = recurse

    if obj_data.n_src == 0:
        print("Empty object", obj_data.n_src, obj_data.n_unique, recurse)
        return

    assert obj_data.data is not None
    if obj_data.mask.sum() == 1:
        obj_data.x_cent = obj_data.data.x_pix.values[0]
        obj_data.y_cent = obj_data.data.y_pix.values[0]
        obj_data.dist_2 = np.zeros((1), float)
        obj_data.rms_dist = 0.0
        obj_data.snr_mean = obj_data.data.snr.values[0]
        obj_data.snr_rms = 0.0
        return

    sum_snr = np.sum(obj_data.data.snr)
    obj_data.x_cent = np.sum(obj_data.data.x_pix * obj_data.data.snr) / sum_snr
    obj_data.y_cent = np.sum(obj_data.data.y_pix * obj_data.data.snr) / sum_snr
    obj_data.dist_2 = np.array(
        (obj_data.x_cent - obj_data.data.x_pix) ** 2
        + (obj_data.y_cent - obj_data.data.y_pix) ** 2
    )
    obj_data.rms_dist = np.sqrt(np.mean(obj_data.dist_2))
    obj_data.snr_mean = np.mean(obj_data.data.snr.values)
    obj_data.snr_rms = np.std(obj_data.data.snr.values)

    sub_mask = obj_data.dist_2 < pixel_r2_cut
    if sub_mask.all():
        return

    if recurse >= cell_data.matcher.max_sub_division:
        return

    heirarchicalSplitObject(obj_data, cell_data, pixel_r2_cut, recurse=recurse + 1)
    return


def heirarchicalSplitObject(
    obj_data: ObjectData, cell_data: CellData, pixel_r2_cut: float, recurse: int = 0
) -> None:
    """Split up a cluster keeping only one source per input
    catalog

    Parameters
    ----------
    obj_data:
        Object being processed

    cell_data:
        Associated cell

    pixel_r2_cut:
        Distance cut for associated, in pixels**2

    recurse:
        Current recursion level

    Notes
    -----
    This function actually does the splitting of objects
    """

    if recurse > cell_data.matcher.max_sub_division:
        return
    obj_data.recurse = recurse
    obj_data.extract()

    assert obj_data.data is not None

    slice_x = obj_data.parent_cluster.footprint.slice_x
    slice_y = obj_data.parent_cluster.footprint.slice_y
    zoom_factors = [1, 2, 4, 8, 16, 32]
    zoom_factor = zoom_factors[recurse]

    n_pix = zoom_factor * np.array(
        [slice_x.stop - slice_x.start, slice_y.stop - slice_y.start]
    )
    zoom_x = (
        zoom_factor
        * obj_data.data.x_cluster
        / obj_data.parent_cluster.pixel_match_scale
    )
    zoom_y = (
        zoom_factor
        * obj_data.data.y_cluster
        / obj_data.parent_cluster.pixel_match_scale
    )

    counts_map = utils.fillCountsMapFromArrays(zoom_x, zoom_y, n_pix)

    fp_dict = utils.getFootprints(counts_map, buf=0)
    footprints = fp_dict["footprints"]
    n_footprints = len(footprints.footprints)
    if n_footprints == 1:
        if recurse >= cell_data.matcher.max_sub_division:
            return
        heirarchicalSplitObject(obj_data, cell_data, pixel_r2_cut, recurse=recurse + 1)
        return

    footprint_key = fp_dict["footprint_key"]
    footprint_ids = utils.findClusterIdsFromArrays(
        zoom_x.values.astype(int), zoom_y.values.astype(int), footprint_key
    )

    biggest = np.argmax(np.bincount(footprint_ids))
    biggest_mask = np.zeros(obj_data.mask.shape, dtype=bool)

    for i_fp in range(n_footprints):
        sub_mask = footprint_ids == i_fp
        count = 0
        new_mask = np.zeros(obj_data.mask.shape, dtype=bool)
        for i, val in enumerate(obj_data.mask):
            if val:
                new_mask[i] = sub_mask[count]
                count += 1
        assert sub_mask.sum() == new_mask.sum()

        if i_fp == biggest:
            biggest_mask = new_mask
            continue

        new_object = obj_data.parent_cluster.addObject(cell_data, new_mask)
        heirarchicalProcessObject(new_object, cell_data, pixel_r2_cut, recurse=recurse)

    obj_data.mask = biggest_mask
    obj_data.extract()
    heirarchicalProcessObject(obj_data, cell_data, pixel_r2_cut, recurse=recurse)
    return


def heirarchicalProcessCluster(
    cluster: ClusterData, cell_data: CellData, pixel_r2_cut: float
) -> list[ObjectData]:
    """Function that is called recursively to
    split clusters until they consist only of sources within
    the match radius of the cluster centroid.

    Recursively process a cluster and make associated objects

    Parameters
    ----------
    cluster:
        Cluster being processed

    cell_data:
        Associated cell

    pixel_r2_cut:
        Distance cut for associated, in pixels**2

    Returns
    -------
    Objects produced in cluster processing
    """
    cluster.extract(cell_data)
    assert cluster.data is not None

    cluster.n_src = len(cluster.data.i_cat)
    cluster.n_unique = len(np.unique(cluster.data.i_cat.values))
    if cluster.n_src == 0:
        print("Empty cluster", cluster.n_src, cluster.n_unique)
        return cluster.objects

    if cluster.n_src == 1:
        cluster.x_cent = cluster.data.x_pix.values[0]
        cluster.y_cent = cluster.data.y_pix.values[0]
        cluster.dist_2 = np.zeros((1))
        cluster.rms_dist = 0.0
        cluster.snr_mean = cluster.data.snr.values[0]
        cluster.snr_rms = 0.0
        initial_object = cluster.addObject(cell_data)
        heirarchicalProcessObject(initial_object, cell_data, pixel_r2_cut)
        return cluster.objects

    sum_snr = np.sum(cluster.data.snr)
    cluster.x_cent = np.sum(cluster.data.x_pix * cluster.data.snr) / sum_snr
    cluster.y_cent = np.sum(cluster.data.y_pix * cluster.data.snr) / sum_snr
    cluster.dist_2 = (cluster.x_cent - cluster.data.x_pix) ** 2 + (
        cluster.y_cent - cluster.data.y_pix
    ) ** 2
    cluster.rms_dist = np.sqrt(np.mean(cluster.dist_2))
    cluster.snr_mean = np.mean(cluster.data.snr.values)
    cluster.snr_rms = np.std(cluster.data.snr.values)

    initial_object = cluster.addObject(cell_data)
    heirarchicalProcessObject(initial_object, cell_data, pixel_r2_cut)
    return cluster.objects



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
import pandas
import tables_io

from .shear_data import ShearData

if TYPE_CHECKING:
    from .cell import CellData
    from .shear_match import ShearMatch


# These are the names of the various shear catalogs
# They are in reverse "alphabetic" order
SHEAR_NAMES = ["ns", "2p", "2m", "1p", "1m"]

# These are the coeffs for the various shear catalogs
DESHEAR_COEFFS = np.array(
    [
        [0, 0, 0, 0],
        [0, 1, 1, 0],
        [0, -1, -1, 0],
        [1, 0, 0, -1],
        [-1, 0, 0, 1],
    ]
)

# These parameters will have to change if the cells change
CELL_INNER_SIZE = 150
CELL_BUFFER = 25
CELL_OFFSET = 0.5
N_PATCH = 20

# These are calculated from the above
CELL_OUTER_SIZE = CELL_INNER_SIZE + (2 * CELL_BUFFER)
CLEAN_CELL_CUT = int(CELL_INNER_SIZE) / 2
UNCLEAN_CELL_CUT = CLEAN_CELL_CUT + 5
PATCH_OFFSET = (N_PATCH + 1 / 2)


def shearStats(df: pandas.DataFrame) -> dict:
    """Return the shear statistics

    {st} is the shear type, one of "gauss", "pgauss", "wmom"

    {i}, {j} index the shear parameters 1, 2

    Parameters
    ----------
    df:
        Input DataFrame, must have :py:class:`hpmcm.ShearTable` schema

    Returns
    -------
    Shear stats in a dict.

    Notes
    -----
    Shear stats include:

    +-----------------+-----------------------------------------------------+
    | Key             | Description                                         |
    +=================+=====================================================+
    | n_{st}          | Number of sources from that catalog                 |
    +-----------------+-----------------------------------------------------+
    | g_{i}_{st}      | g_{i} shear parameter for that catalog              |
    +-----------------+-----------------------------------------------------+
    | delta_g_{i}_{j} | g_{i,j} shear measurment: g_{i}_{j}p - g_{i}_{j}m   |
    +-----------------+-----------------------------------------------------+
    | good            | True if every catalog has one source in this object |
    +-----------------+-----------------------------------------------------+

    If the matching is not good, then delta_g_1 = delta_g_2 = np.nan
    """
    out_dict: dict[str, float | int] = {}
    all_good = True
    for i, name_ in enumerate(SHEAR_NAMES):
        mask = df.i_cat == i
        n_cat = mask.sum()
        if n_cat != 1:
            all_good = False
        out_dict[f"n_{name_}"] = int(n_cat)
        if n_cat:
            out_dict[f"g_1_{name_}"] = df[mask].g_1.values.mean()
            out_dict[f"g_2_{name_}"] = df[mask].g_2.values.mean()
        else:
            out_dict[f"g_1_{name_}"] = np.nan
            out_dict[f"g_2_{name_}"] = np.nan
    if all_good:
        out_dict["delta_g_1_1"] = out_dict["g_1_1p"] - out_dict["g_1_1m"]
        out_dict["delta_g_2_2"] = out_dict["g_2_2p"] - out_dict["g_2_2m"]
        out_dict["delta_g_1_2"] = out_dict["g_1_2p"] - out_dict["g_1_2m"]
        out_dict["delta_g_2_1"] = out_dict["g_2_1p"] - out_dict["g_2_1m"]
    else:
        out_dict["delta_g_1_1"] = np.nan
        out_dict["delta_g_2_2"] = np.nan
        out_dict["delta_g_1_2"] = np.nan
        out_dict["delta_g_2_1"] = np.nan
    out_dict["good"] = all_good
    return out_dict


def shearReport(
    basefile: str,
    output_file_base: str | None,
    shear: float,
    cat_type: str,
    tract: int,
    snr_cut: float = 7.5,
) -> None:
    """Report on the shear calibration

    Parameters
    ----------
    basefile
        Input base file name (see notes)

    output_file_base:
        Output file name (see notes)

    shear:
        Applied shear

    cat_type:
        Catalog type (one of ["pgauss", "gauss", "wmom"]

    tract:
        Tract, written to outout data

    snr_cut:
        Signal-to-noise cut.

    Notes
    -----
    This will read the object shear data from "{basefile}_cluster_shear.pq"
    This will read the object statisticis from "{basefile}_cluster_stats.pq"

    This will write the shear stats to "{output_file_base}.pkl"
    This will write the figures to "{output_file_base}_{figure}.png"
    """
    t = tables_io.read(f"{basefile}_cluster_shear.pq")
    t2 = tables_io.read(f"{basefile}_cluster_stats.pq")

    shear_data = ShearData(t, t2, shear, cat_type, tract, snr_cut=snr_cut)

    if output_file_base is not None:
        shear_data.save(f"{output_file_base}.pkl")
        shear_data.savefigs(output_file_base)


def mergeShearReports(
    inputs: list[str],
    output_file: str,
) -> None:
    """Merge reports on the shear calibration

    Parameters
    ----------
    inputs:
        List of input ShearData pickle files

    output_file:
        Where to write the merged file
    """
    out_dict: dict[str, Any] = {}
    for input_ in inputs:
        shear_data = ShearData.load(input_)
        input_dict = shear_data.toDict()
        for key, val in input_dict.items():
            if key in out_dict:
                out_dict[key].append(val)
            else:
                out_dict[key] = [val]

    out_df = pandas.DataFrame(out_dict)
    out_df.to_parquet(output_file)


def splitByTypeAndClean(
    basefile: str,
    tract: int,
    shear: float,
    cat_type: str,
    *,
    clean: bool = False,
) -> None:  # pragma: no cover
    """Split a parquet file by shear catalog type and tract

    Parameters
    ----------
    basefile:
        Original file name

    tract:
        Tract to select

    shear:
        Applied shear, saved to output

    cat_type:
        Catalog type to select

    clean:
        Remove duplicates

    Notes
    -----
    This will create 5 files with the pattern:
    "{basefile}_uncleaned_{tract}_{type}.pq"

    +--------------+-------------------------------------+
    | Column       | Description                         |
    +==============+=====================================+
    | id           | Index of object inside catalog      |
    +--------------+-------------------------------------+
    | orig_id      | Original object id                  |
    +--------------+-------------------------------------+
    | cell_idx_x   | X-index of Cell                     |
    +--------------+-------------------------------------+
    | cell_idx_y   | Y-index of Cell                     |
    +--------------+-------------------------------------+
    | x_cell_coadd | X-coordinate in cell frame          |
    +--------------+-------------------------------------+
    | y_cell_coadd | Y-coordinate in cell frame          |
    +--------------+-------------------------------------+
    | x_pix        | X-coordinate in global WCS frame    |
    +--------------+-------------------------------------+
    | y_pix        | Y-coordinate in global WCS frame    |
    +--------------+-------------------------------------+
    | g_1          | Shear g_1 component estimate        |
    +--------------+-------------------------------------+
    | g_2          | Shear g_2 component estimate        |
    +--------------+-------------------------------------+
    | snr          | Signal-to-noise ratio               |
    +--------------+-------------------------------------+

    """
    p = tables_io.read(basefile)
    if clean:
        clean_st = "cleaned"
        cell_cut = CLEAN_CELL_CUT
    else:
        clean_st = "uncleaned"
        cell_cut = UNCLEAN_CELL_CUT
    for type_ in SHEAR_NAMES:
        mask = p["shear_type"] == type_
        sub = p[mask].copy(deep=True)
        cell_idx_x = (N_PATCH * sub["patch_x"].values + sub["cell_x"].values).astype(
            int
        )
        cell_idx_y = (N_PATCH * sub["patch_y"].values + sub["cell_y"].values).astype(
            int
        )
        cent_x = CELL_INNER_SIZE * (cell_idx_x - CELL_OFFSET)
        cent_y = CELL_INNER_SIZE * (cell_idx_y - CELL_OFFSET)
        x_cell_coadd = sub["col"] - cent_x
        y_cell_coadd = sub["row"] - cent_y
        sub["x_pix"] = sub["col"] + CELL_BUFFER
        sub["y_pix"] = sub["row"] + CELL_BUFFER
        sub["x_cell_coadd"] = x_cell_coadd
        sub["y_cell_coadd"] = y_cell_coadd
        sub["snr"] = sub[f"{cat_type}_band_flux_r"] / sub[f"{cat_type}_band_flux_err_r"]
        sub["g_1"] = sub[f"{cat_type}_g_1"]
        sub["g_2"] = sub[f"{cat_type}_g_2"]
        sub["cell_idx_x"] = cell_idx_x
        sub["cell_idx_y"] = cell_idx_y
        sub["orig_id"] = sub.id
        sub["id"] = np.arange(len(sub))
        central_to_cell = np.bitwise_and(
            np.fabs(x_cell_coadd) < cell_cut, np.fabs(y_cell_coadd) < cell_cut
        )
        central_to_patch = np.bitwise_and(
            np.fabs(sub["cell_x"].values - PATCH_OFFSET) < (N_PATCH/2),
            np.fabs(sub["cell_y"].values - PATCH_OFFSET) < (N_PATCH/2)
        )
        right_tract = sub["tract"] == tract
        central = np.bitwise_and(central_to_cell, central_to_patch)
        selected = np.bitwise_and(right_tract, central)
        cleaned = sub[selected].copy(deep=True)
        cleaned["shear"] = np.repeat(shear, len(cleaned))
        cleaned.to_parquet(basefile.replace(".parq", f"_{clean_st}_{tract}_{type_}.pq"))


def reduceShearDataForCell(
    cell: CellData, i_cat: int, dataframe: pandas.DataFrame
) -> pandas.DataFrame:
    """Filters dataframe to keep only sources in the cell

    Parameters
    ----------
    cell:
        The cell being analyzed

    i_cat:
        Catalog index

    dataframe:
        Input dataframe


    Returns
    -------
    Filtered datasets


    Notes
    -----
    This will optionally deshear the source positions if `matcher.deshear`
    is not None.

    This will add these columns to the output dataframes

    +-----------+-------------------------------------+
    | Column    | Description                         |
    +===========+=====================================+
    | x_cell    | X-coordinate in cell frame          |
    +-----------+-------------------------------------+
    | y_cell    | Y-coordinate in cell frame          |
    +-----------+-------------------------------------+
    | x_pix     | X-coordinate in global WCS frame    |
    +-----------+-------------------------------------+
    | y_pix     | Y-coordinate in global WCS frame    |
    +-----------+-------------------------------------+
    | dx_shear  | Change in X position when desheared |
    +-----------+-------------------------------------+
    | dy_shear  | Change in X position when desheared |
    +-----------+-------------------------------------+

    """

    matcher = cell.matcher

    if TYPE_CHECKING:
        assert isinstance(matcher, ShearMatch)

    filtered_idx = matcher.getCellIndices(dataframe) == cell.idx
    reduced = dataframe[filtered_idx].copy(deep=True)

    x_cell_orig = reduced["x_cell_coadd"]
    y_cell_orig = reduced["y_cell_coadd"]
    x_pix_orig = reduced["x_pix"]
    y_pix_orig = reduced["y_pix"]

    if matcher.deshear is not None:
        # De-shear in the cell frame to do matching
        dx_shear = matcher.deshear * (
            x_cell_orig * DESHEAR_COEFFS[i_cat][0]
            + y_cell_orig * DESHEAR_COEFFS[i_cat][2]
        )
        dy_shear = matcher.deshear * (
            x_cell_orig * DESHEAR_COEFFS[i_cat][1]
            + y_cell_orig * DESHEAR_COEFFS[i_cat][3]
        )
        x_cell = x_cell_orig + dx_shear
        y_cell = y_cell_orig + dy_shear
        x_pix = x_pix_orig + dx_shear
        y_pix = y_pix_orig + dy_shear
    else:  # pragma: no cover
        dx_shear = np.zeros(len(dataframe))
        dy_shear = np.zeros(len(dataframe))
        x_cell = x_cell_orig
        y_cell = y_cell_orig
        x_pix = x_pix_orig
        y_pix = y_pix_orig

    x_cell = (x_cell + (CELL_OUTER_SIZE/2)) / matcher.pixel_match_scale
    y_cell = (y_cell + (CELL_OUTER_SIZE/2)) / matcher.pixel_match_scale
    filtered_x = np.bitwise_and(x_cell >= 0, x_cell < cell.n_pix[0])
    filtered_y = np.bitwise_and(y_cell >= 0, y_cell < cell.n_pix[1])
    filtered_bounds = np.bitwise_and(filtered_x, filtered_y)
    red = reduced[filtered_bounds].copy(deep=True)
    red["x_cell"] = x_cell[filtered_bounds]
    red["y_cell"] = y_cell[filtered_bounds]
    red["x_pix"] = x_pix[filtered_bounds]
    red["y_pix"] = y_pix[filtered_bounds]
    if matcher.deshear is not None:
        red["dx_shear"] = dx_shear[filtered_bounds]
        red["dy_shear"] = dy_shear[filtered_bounds]
    return red


def makeMatchedShearSourceCatalogs(
    source_base_name: str,
    match_base_name: str,
) -> dict[str, pandas.DataFrame]:
    """Use the associations to join the source tables to their match obects

    Parameters
    ----------
    source_base_name:
        _base file name for souces catalogs

    match_base_name:
        _base file naem for match tables

    Returns
    -------
    Dict of tables, keyed by shear type, which have the
    souces catalogs joined to the associated objects
    """
    keys = ["object_stats", "object_assoc", "object_shear"]
    shear_types = {v: k for k, v in enumerate(SHEAR_NAMES)}
    td = tables_io.read(match_base_name, keys=keys)
    itd = tables_io.read(source_base_name, keys=list(shear_types.keys()))
    td["object_stats"]["idx"] = np.arange(len(td["object_stats"]))
    td["object_shear"]["idx"] = np.arange(len(td["object_shear"]))
    merged_object = td["object_stats"].merge(
        td["object_shear"], on="idx", how="inner", suffixes=["_l", "_r"]
    )
    merged_object_assoc = td["object_assoc"].merge(
        merged_object, on="objectId", how="inner", suffixes=["_l", "_r"]
    )
    out_dict = {}
    for cat_type_, i_cat_ in shear_types.items():
        merged_object_assoc_mask = merged_object_assoc.catalogId == i_cat_
        merged_object_assoc_masked = merged_object_assoc[merged_object_assoc_mask]
        sources = itd[cat_type_]
        sources["sourceId"] = sources["id"]
        matched_source = merged_object_assoc_masked.merge(
            sources, on="sourceId", how="inner", suffixes=["_l", "_r"]
        )
        out_dict[cat_type_] = matched_source
    return out_dict



import enum
from functools import partial
from typing import Any, Type, TypeVar

import click

__all__: list[str] = [
    "basefile",
    "catalog_type",
    "deshear",
    "inputs",
    "output_file",
    "cell_size",
    "ra_ref",
    "dec_ref",
    "ra_size",
    "dec_size",
    "pixel_size",
    "pixel_match_scale",
    "pixel_r2_cut",
    "shear",
    "snr_cut",
    "tract",
]

EnumType_co = TypeVar(  # pylint: disable=invalid-name
    "EnumType_co", bound=Type[enum.Enum], covariant=True
)


class EnumChoice(click.Choice):  # pragma: no cover
    """A version of click.Choice specialized for enum types"""

    def __init__(self, the_enum: EnumType_co, case_sensitive: bool = True) -> None:
        self._enum = the_enum
        super().__init__(
            list(the_enum.__members__.keys()), case_sensitive=case_sensitive
        )

    def convert(
        self, value: Any, param: click.Parameter | None, ctx: click.Context | None
    ) -> enum.Enum:
        converted_str = super().convert(value, param, ctx)
        return self._enum.__members__[converted_str]


class PartialOption:
    """Wraps click.option with partial arguments for convenient reuse"""

    def __init__(self, *param_decls: str, **kwargs: Any) -> None:
        self._partial = partial(click.option, *param_decls, cls=click.Option, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._partial(*args, **kwargs)


class PartialArgument:
    """Wraps click.argument with partial arguments for convenient reuse"""

    def __init__(self, *param_decls: Any, **kwargs: Any) -> None:
        self._partial = partial(
            click.argument, *param_decls, cls=click.Argument, **kwargs
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:  # pragma: no cover
        return self._partial(*args, **kwargs)


basefile = PartialOption(
    "--basefile",
    type=click.Path(),
    help="Input file basename",
    required=True,
)

catalog_type = PartialOption(
    "--catalog-type",
    help="Which type of catalog 'gauss', 'pgauss', 'wmom'",
    type=str,
)

deshear = PartialOption(
    "--deshear",
    help="Deshear postions before match",
    is_flag=True,
)

output_file = PartialOption(
    "--output-file",
    type=click.Path(),
    help="Path for output file",
    required=True,
)


ra_ref = PartialOption(
    "--ra-ref",
    help="Reference RA coordinate for WCS",
    type=float,
)


dec_ref = PartialOption(
    "--dec-ref",
    help="Reference DEC coordinate for WCS",
    type=float,
)


ra_size = PartialOption(
    "--ra-size",
    help="Size of match region along RA direction",
    type=float,
)


dec_size = PartialOption(
    "--dec-size",
    help="Size of match region along DEC direction",
    type=float,
)


cell_size = PartialOption(
    "--cell-size",
    help="Size of cell in pixels",
    type=int,
    default=1000,
)


pixel_size = PartialOption(
    "--pixel-size",
    help="Size of pixels for WCS matching maps",
    type=float,
)


pixel_r2_cut = PartialOption(
    "--pixel-r2-cut",
    help="Matching cut to apply in pixel distance squared",
    type=float,
    default=4.0,
)

pixel_match_scale = PartialOption(
    "--pixel-match-scale",
    help="Scale factor to apply for pixel in matching",
    type=int,
    default=1,
)

snr_cut = PartialOption(
    "--snr-cut",
    help="signal-to-noise cut",
    type=float,
    default=7.5,
)


shear = PartialOption(
    "--shear",
    help="Shear value",
    type=float,
    default=None,
)

tract = PartialOption("--tract", help="Tract to select", type=int, required=True)

inputs = PartialArgument("inputs", nargs=-1)



import click
import numpy as np
import tables_io

import hpmcm
from hpmcm import __version__

from . import options

__all__ = [
    "cli",
    "wcsGroup",
    "wcsMatchCommand",
    "shearGroup",
    "shearMatchCommand",
    "shearSplitCommand",
    "shearReportCommand",
]


@click.group()
@click.version_option(__version__)
def cli() -> None:
    """HPMCM Command line interface"""


@cli.group(name="wcs")
def wcsGroup() -> None:
    """Operations on catalogs using single WCS"""


@wcsGroup.command(name="match")
@options.inputs()
@options.output_file()
@options.ra_ref(required=True)
@options.dec_ref(required=True)
@options.ra_size(required=True)
@options.dec_size(required=True)
@options.pixel_size(required=True)
@options.cell_size()
@options.pixel_r2_cut()
def wcsMatchCommand(
    inputs: list[str],
    output_file: click.Path,
    ra_ref: float,
    dec_ref: float,
    ra_size: float,
    dec_size: float,
    pixel_size: float,
    cell_size: int,
    pixel_r2_cut: float | None,
) -> None:
    """Match catalogs using a global WCS map"""
    matcher = hpmcm.WcsMatch.create(
        ref_dir=(ra_ref, dec_ref),
        region_size=(ra_size, dec_size),
        pixel_size=pixel_size,
        cell_size=cell_size,
        pixel_R2_cut=pixel_r2_cut,
    )
    input_files = list(inputs)
    visit_ids: list[int] = list(np.arange(len(input_files)))
    matcher.reduceData(input_files, visit_ids)
    print("Matching catalogs:")
    matcher.analysisLoop()
    print("Extracting stats.")
    stats = matcher.extractStats()
    print("Writing output.")
    out_dict = dict(
        _cluster_assoc=stats[0],
        _object_assoc=stats[1],
        _cluster_stats=stats[2],
        _object_stats=stats[3],
    )
    tables_io.write(out_dict, output_file)
    print("Success!")


@cli.group(name="shear")
def shearGroup() -> None:
    """Operations on shear catalogs"""


@shearGroup.command(name="match")
@options.inputs()
@options.output_file()
@options.shear()
@options.pixel_r2_cut()
@options.pixel_match_scale()
@options.deshear()
def shearMatchCommand(
    inputs: list[str],
    output_file: click.Path,
    shear: float | None,
    pixel_r2_cut: float | None,
    pixel_match_scale: int,
    deshear: bool,
) -> None:
    """Match shear catalogs"""
    deshear_value: float | None = None
    if deshear:
        assert shear is not None
        deshear_value = -1 * shear
    matcher = hpmcm.ShearMatch.createShearMatch(
        pixelR2Cut=pixel_r2_cut,
        pixelMatchScale=pixel_match_scale,
        deshear=deshear_value,
    )
    input_files = list(inputs)
    input_files.reverse()
    visit_ids: list[int] = list(np.arange(len(input_files)))
    matcher.reduceData(input_files, visit_ids)
    print("Matching catalogs:")
    matcher.analysisLoop()
    print("Extracting stats.")
    stats = matcher.extractStats()
    print("Extracting shear stats.")
    shear_stats = matcher.extractShearStats()
    print("Writing output.")
    out_dict = dict(
        _cluster_assoc=stats[0],
        _object_assoc=stats[1],
        _cluster_stats=stats[2],
        _object_stats=stats[3],
        _cluster_shear=shear_stats[0],
        _object_shear=shear_stats[1],
    )
    tables_io.write(out_dict, output_file)
    print("Success!")


@shearGroup.command(name="split")
@options.basefile()
@options.tract()
@options.catalog_type()
@options.shear(required=True)
def shearSplitCommand(
    basefile: str,
    tract: int,
    shear: float,
    catalog_type: str,
) -> None:
    """Split input shear catalogs"""
    hpmcm.shear_utils.splitByTypeAndClean(basefile, tract, shear, catalog_type)


@shearGroup.command(name="report")
@options.basefile()
@options.output_file()
@options.shear(required=True)
@options.snr_cut()
def shearReportCommand(
    basefile: str,
    output_file: str,
    shear: float,
    snr_cut: float,
) -> None:
    """Build shear calibration reports"""
    tokens = basefile.split("_")
    tract = int(tokens[-1])
    cat_type = tokens[-4]
    hpmcm.shear_utils.shearReport(
        basefile, output_file, shear, cat_type=cat_type, tract=tract, snr_cut=snr_cut
    )


@shearGroup.command(name="merge-reports")
@options.inputs()
@options.output_file()
def shearMergeReportsCommand(
    inputs: list[str],
    output_file: str,
) -> None:
    """Build shear calibration reports"""
    hpmcm.shear_utils.mergeShearReports(inputs, output_file)



import os
import subprocess
from urllib.request import urlretrieve


def setupTestDataArea() -> int:  # pragma: no cover
    """Download test files to setup a project testsing area

    Returns
    -------
    int:
       0 for success, error code otherwise

    Notes
    -----
    This will download files into 'examples/test_data', and could take a few
    minutes.

    This will not download the files if they are already present
    """
    if not os.path.exists("examples/test_data"):

        if not os.path.exists("examples/test_data.tgz"):
            urlretrieve(
                "http://s3df.slac.stanford.edu/people/echarles/package_test_data/hpmcm/hpmcm_test_data.tgz",
                "examples/test_data.tgz",
            )
            if not os.path.exists("examples/test_data.tgz"):
                return 1

        status = subprocess.run(
            ["tar", "zxvf", "examples/test_data.tgz", "-C", "examples"], check=False
        )
        if status.returncode != 0:
            return status.returncode

    if not os.path.exists("examples/test_data/object_10463.pq"):
        return 2

    return 0


def teardownTestDataArea() -> None:  # pragma: no cover
    """Remove the test data

    Notes
    -----
    If the env var "NO_TEARDOWN" is set, this will not remove the data
    """
    if "NO_TEARDOWN" in os.environ:
        return
    os.system("\\rm -rf examples/test_data")
    try:
        os.unlink("examples/test_data.tgz")
    except FileNotFoundError:
        pass



from __future__ import annotations

from typing import Any

import numpy as np
import pandas
from astropy import wcs

from .match import Match


def createGlobalWcs(
    ref_dir: tuple[float, float],
    pix_size: float,
    n_pix: np.ndarray,
) -> wcs.WCS:
    """Helper function to create the WCS used to project the
    sources in a skymap


    Parameters
    ----------
    ref_dir:
        Reference Direction (RA, DEC) in degrees

    pix_size:
        Pixel size in degrees

    n_pix:
        Number of pixels in x, y

    Returns
    -------
    WCS to create the pixel grid
    """
    w = wcs.WCS(naxis=2)
    w.wcs.cdelt = [-pix_size, pix_size]
    w.wcs.crpix = [n_pix[0] / 2, n_pix[1] / 2]
    w.wcs.crval = [ref_dir[0], ref_dir[1]]
    return w


class WcsMatch(Match):
    """Class to do N-way matching using a provided WCS.

    This uses the provided WCS to define a Skymap that covers the full region
    begin matched.

    Notes
    -----
    This expectes a list of parquet files with pandas DataFrames
    that contain the following columns.

    +-------------+---------------------------------------------------------------+
    | Column name | Description                                                   |
    +=============+===============================================================+
    | id          | source ID                                                     |
    +-------------+---------------------------------------------------------------+
    | ra          | RA in degrees                                                 |
    +-------------+---------------------------------------------------------------+
    | dec         | DEC in degress                                                |
    +-------------+---------------------------------------------------------------+
    | snr         | Signal-to-Noise of source, used for filtering and centroiding |
    +-------------+---------------------------------------------------------------+
    """

    def __init__(
        self,
        match_wcs: wcs.WCS,
        **kwargs: Any,
    ):
        self.wcs: wcs.WCS = match_wcs
        pixel_size: float = self.wcs.wcs.cdelt[1]
        n_pix_side: np.ndarray = np.ceil(2 * np.array(self.wcs.wcs.crpix)).astype(int)
        Match.__init__(self, pixel_size=pixel_size, n_pix_side=n_pix_side, **kwargs)

    @classmethod
    def create(
        cls,
        ref_dir: tuple[float, float],
        region_size: tuple[float, float],
        pixel_size: float,
        **kwargs: Any,
    ) -> WcsMatch:
        """Helper function to create a Match object.

        This will use the provided arguments to create a WCS, and then
        create and return a `WcsMatch` matcher.

        Parameters
        ----------
        ref_dir:
            Reference Direction (RA, DEC) in degrees

        region_size:
            Size of match region, in degrees

        pix_size:
            Pixel size in degrees

        Returns
        -------
        Object to create matches for the requested region
        """
        n_pix = (np.array(region_size) / pixel_size).astype(int)
        match_wcs = createGlobalWcs(ref_dir, pixel_size, n_pix)
        return cls(match_wcs, n_pixels=n_pix, **kwargs)

    def _getPixValues(self, df: pandas.DataFrame) -> tuple[np.ndarray, np.ndarray]:
        x_pix, y_pix = self.wcs.wcs_world2pix(df["ra"].values, df["dec"].values, 0)
        return x_pix, y_pix

    def pixToWorld(
        self,
        x_pix: np.ndarray,
        y_pix: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Convert local coords in pixels to world coordinates (RA, DEC)"""
        assert self.wcs is not None
        return self.wcs.wcs_pix2world(x_pix, y_pix, 0)

    def reduceDataFrame(
        self,
        df: pandas.DataFrame,
    ) -> pandas.DataFrame:
        """Reduce a single input DataFrame

        Notes
        -----
        This applies a trivial cut on signal-to-noise (snr>1).
        and adds the pixel coordinates to the DataFrame.

        This will add these columns to the output dataframes


        +--------------+-------------------------------------+
        | Column       | Description                         |
        +==============+=====================================+
        | id           | Index of object inside catalog      |
        +--------------+-------------------------------------+
        | ra           | Source RA                           |
        +--------------+-------------------------------------+
        | dec          | Source DEC                          |
        +--------------+-------------------------------------+
        | x_pix        | X-coordinate in global WCS frame    |
        +--------------+-------------------------------------+
        | y_pix        | Y-coordinate in global WCS frame    |
        +--------------+-------------------------------------+
        | snr          | Signal-to-noise ratio               |
        +--------------+-------------------------------------+

        """
        df_clean = df[(df.snr > 1)]
        x_pix, y_pix = self.wcs.wcs_world2pix(
            df_clean["ra"].values, df_clean["dec"].values, 0
        )
        df_clean["x_pix"] = x_pix
        df_clean["y_pix"] = y_pix
        filtered = (
            (df_clean.x_pix >= 0)
            & (df_clean.x_pix < self.n_pix_side[0])
            & (df_clean.y_pix >= 0)
            & (df_clean.y_pix < self.n_pix_side[1])
        )

        df_red = df_clean[filtered].copy(deep=True)

        return df_red[
            [
                "id",
                "ra",
                "dec",
                "x_pix",
                "y_pix",
                "snr",
            ]
        ]



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING

import numpy as np
import pandas

from . import match_utils, shear_utils, utils
from .cluster import ClusterData, ShearClusterData
from .footprint import Footprint, FootprintSet
from .object import ObjectData, ShearObjectData

if TYPE_CHECKING:
    from .match import Match
    from .shear_match import ShearMatch


class CellData:
    """Class to store analyze data for a cell

    Includes cell boundries, reduced data tables
    and clustering results

    Does not store sky maps

    Cells are square sub-regions of the analysis region
    that are extracted from the ranges of pixels in the WCS

    The cell covers corner:corner+size

    The sources are projected into an array that extends `buf` pixels
    beyond the cell

    This uses the FootprintSet to identify pixels which contain
    source, and builds those into clusters

    Attributes
    ----------
    matcher: Match
        Parent Match object

    id_offset: int
        Offset used for the Object and Cluster IDs for this cell

    corner: np.ndarray
        pixX, pixY for corner of cell

    size: np.ndarray
        size of the cell (in pixels)

    idx: int
        Id of the cell

    buf: int
        Number of buffer pixels around the edge of the cell

    min_pix: np.ndarray
        Lowest number pixel in center of cell

    max_pix: np.ndarray
        Highest number pixel in center of cell

    n_pix: np.ndarray
        Number of pixels in center of cell

    data : list[pandas.DataFrame]
        Reduced dataframes with only sources for this cell

    n_src : int
        Number of sources in this cell

    footprint_ids : list[np.ndarray]
        Matched arrays with the index of the cluster associated to each
        source.  I.e., these could added to the Dataframes as
        additional columns

    cluster_dict : OrderedDict[int, ClusterData]
        Dictionary with cluster membership data

    object_dict : OrderedDict[int, ObjectData]
        Dictionary with object membership data

    """

    def __init__(
        self,
        matcher: Match,
        id_offset: int,
        corner: np.ndarray,
        size: np.ndarray,
        idx: int,
        buf: int = 10,
    ):
        self.matcher: Match = matcher
        # Offset used for the Object and Cluster IDs for this cell
        self.id_offset: int = id_offset
        self.corner: np.ndarray = corner  # pixX, pixY for corner of cell
        self.size: np.ndarray = size  # size of cell
        self.idx: int = idx  # cell index
        self.buf: int = buf
        self.min_pix: np.ndarray = corner - buf
        self.max_pix: np.ndarray = corner + size + buf
        self.n_pix: np.ndarray = self.max_pix - self.min_pix

        self.data: list[pandas.DataFrame] = []
        self.n_src: int = 0
        self.footprint_ids: list[np.ndarray] = []
        self.cluster_dict: OrderedDict[int, ClusterData] = OrderedDict()
        self.object_dict: OrderedDict[int, ObjectData] = OrderedDict()

    def reduceData(self, data: list[pandas.DataFrame]) -> None:
        """Pull out only the data needed for this cell"""
        self.data = [self.reduceDataframe(i, val) for i, val in enumerate(data)]
        self.n_src = int(np.sum([len(df) for df in self.data]))

    @property
    def n_clusters(self) -> int:
        """Return the number of clusters in this cell"""
        return len(self.cluster_dict)

    @property
    def n_objects(self) -> int:
        """Return the number of objects in this cell"""
        return len(self.object_dict)

    def reduceDataframe(
        self, i_cat: int, dataframe: pandas.DataFrame
    ) -> pandas.DataFrame:
        """Filters dataframe to keep only source in the cell"""
        assert i_cat is not None

        # WCS is defined, use it
        x_cell = dataframe["x_pix"] - self.min_pix[0]
        y_cell = dataframe["y_pix"] - self.min_pix[1]
        filtered = (
            (x_cell >= 0)
            & (x_cell < self.n_pix[0])
            & (y_cell >= 0)
            & (y_cell < self.n_pix[1])
        )
        red = dataframe[filtered].copy(deep=True)
        red["x_cell"] = x_cell[filtered]
        red["y_cell"] = y_cell[filtered]
        return red

    def countsMap(self, weight_name: str | None = None) -> np.ndarray:
        """Fill a map that counts the number of source per cell"""
        to_fill = self._emtpyCountsMaps()
        assert self.data is not None
        for df in self.data:
            to_fill += self._singleCatalogCountsMap(df, weight_name)
        return to_fill

    def buildClusterData(
        self,
        fp_set: FootprintSet,
        pixel_r2_cut: float = 4.0,
    ) -> None:
        """Loop through cluster ids and collect sources into
        the ClusterData objects"""
        footprint_dict: dict[int, list[tuple[int, int, int]]] = {}
        n_missing = 0
        n_found = 0
        assert self.data is not None
        assert self.footprint_ids
        for i_cat, (df, footprint_ids) in enumerate(zip(self.data, self.footprint_ids)):
            for src_idx, (src_id, footprint_id) in enumerate(
                zip(df["id"], footprint_ids)
            ):
                if footprint_id < 0:
                    n_missing += 1
                    continue
                if footprint_id not in footprint_dict:
                    footprint_dict[footprint_id] = [(i_cat, src_id, src_idx)]
                else:
                    footprint_dict[footprint_id].append((i_cat, src_id, src_idx))
                n_found += 1
        for footprint_id, sources in footprint_dict.items():
            footprint = fp_set.footprints[footprint_id]
            i_cluster = footprint_id + self.id_offset
            cluster = self._buildClusterData(i_cluster, footprint, np.array(sources).T)
            self.cluster_dict[i_cluster] = cluster
            match_utils.heirarchicalProcessCluster(cluster, self, pixel_r2_cut)

    def analyze(
        self, weight_name: str | None = None, pixel_r2_cut: float = 2.0
    ) -> dict | None:
        """Analyze this cell

        Note that this returns the counts maps and clustering info,
        which can be helpful for debugging.
        """
        if self.n_src == 0:
            return None
        counts_map = self.countsMap(weight_name)
        o_dict = self._getFootprints(counts_map)
        o_dict["counts_map"] = counts_map
        assert self.data is not None
        self.footprint_ids = self._associateSourcesToFootprints(
            self.data, o_dict["footprint_key"]
        )
        self.buildClusterData(o_dict["footprints"], pixel_r2_cut)
        return o_dict

    def addObject(
        self, cluster: ClusterData, mask: np.ndarray | None = None
    ) -> ObjectData:
        """Add an object to this cell

        Parameters
        ----------
        cluster:
            Parent cluster for the object

        mask:
            Mask of which sources in the cluster to include in the object

        Returns
        -------
        Newly created ObjectData
        """
        object_id = self.n_objects + self.id_offset
        new_object = self._newObject(cluster, object_id, mask)
        self.object_dict[object_id] = new_object
        return new_object

    @classmethod
    def _newObject(
        cls, cluster: ClusterData, object_id: int, mask: np.ndarray | None
    ) -> ObjectData:
        return ObjectData(cluster, object_id, mask)

    def _emtpyCountsMaps(self) -> np.ndarray:
        to_fill = np.zeros(np.ceil(self.n_pix).astype(int))
        return to_fill

    def _singleCatalogCountsMap(
        self, df: pandas.DataFrame, weight_name: str | None = None
    ) -> np.ndarray:
        return utils.fillCountsMapFromDf(
            df,
            n_pix=self.n_pix,
            weight_name=weight_name,
        )

    def _buildClusterData(
        self, i_cluster: int, footprint: Footprint, sources: np.ndarray
    ) -> ClusterData:
        return ClusterData(i_cluster, footprint, sources)

    def _getFootprints(self, counts_map: np.ndarray) -> dict:
        return utils.getFootprints(counts_map, buf=self.buf)

    def _associateSourcesToFootprints(
        self,
        data: list[pandas.DataFrame],
        cluster_key: np.ndarray,
    ) -> list[np.ndarray]:
        return utils.associateSourcesToFootprints(
            data,
            cluster_key,
        )

    def getRaDec(
        self, x_cents: np.ndarray, y_cents: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """Return the RA, DEC of based from pixel coords"""
        return self.matcher.pixToWorld(x_cents, y_cents)


class ShearCellData(CellData):
    """Subclass of CellData that can compute shear statisitics

    Attributes
    ----------
    pixel_match_scale: int
        Number of pixel merged in the original counts map
    """

    def __init__(
        self,
        matcher: ShearMatch,
        id_offset: int,
        corner: np.ndarray,
        size: np.ndarray,
        idx: int,
        buf: int = 10,
    ):
        CellData.__init__(self, matcher, id_offset, corner, size, idx, buf)
        self.pixel_match_scale = matcher.pixel_match_scale

    def reduceDataframe(
        self, i_cat: int, dataframe: pandas.DataFrame
    ) -> pandas.DataFrame:
        """Filters dataframe to keep only source in the cell"""
        return shear_utils.reduceShearDataForCell(self, i_cat, dataframe)

    @classmethod
    def _newObject(
        cls, cluster: ClusterData, object_id: int, mask: np.ndarray | None
    ) -> ObjectData:
        return ShearObjectData(cluster, object_id, mask)

    def _emtpyCountsMaps(self) -> np.ndarray:
        pixel_match_scale = self.pixel_match_scale
        to_fill = np.zeros(np.ceil(self.n_pix / pixel_match_scale).astype(int))
        return to_fill

    def _singleCatalogCountsMap(
        self, df: pandas.DataFrame, weight_name: str | None = None
    ) -> np.ndarray:
        return utils.fillCountsMapFromDf(
            df,
            n_pix=self.n_pix,
            weight_name=weight_name,
            pixel_match_scale=self.pixel_match_scale,
        )

    def _buildClusterData(
        self, i_cluster: int, footprint: Footprint, sources: np.ndarray
    ) -> ClusterData:
        return ShearClusterData(
            i_cluster, footprint, sources, pixel_match_scale=self.pixel_match_scale
        )

    def _getFootprints(self, counts_map: np.ndarray) -> dict:
        return utils.getFootprints(
            counts_map, buf=0, pixel_match_scale=self.pixel_match_scale
        )

    def _associateSourcesToFootprints(
        self,
        data: list[pandas.DataFrame],
        cluster_key: np.ndarray,
    ) -> list[np.ndarray]:
        return utils.associateSourcesToFootprints(
            data,
            cluster_key,
            pixel_match_scale=self.pixel_match_scale,
        )

    def getRaDec(
        self, x_cents: np.ndarray, y_cents: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        return np.repeat(np.nan, len(x_cents)), np.repeat(np.nan, len(y_cents))



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas

from . import shear_utils

if TYPE_CHECKING:
    from .cluster import ClusterData


class ObjectData:
    """Small class to define 'Objects', i.e., sets of associated sources

    Attributes
    ----------
    parent_cluster : ClusterData
        Parent cluster that this object is build from

    object_id : int
        Unique object identifier

    mask : np.ndarray[bool]
        Mask of which source in parent cluster are in this object

    catalog_id: np.ndarray[int]
        Indicies showing which catalog each source belongs to

    n_src: int
        Number of sources in this object

    n_unique: int
        Number of catalogs contributing to this object

    data: pandas.DataFrame
        Data the sources in this object

    recurse: int
        Recursion level needed to make this object

    x_cent: float
        X Centeroid of this object in global pixel coordinates

    y_cent: float
        Y Centeroid of this object in global pixel coordinates

    rms_dist: float
        RMS distance of sources to the centroid

    dist_2: np.ndarray
        Distances of sources to the centroid
    """

    def __init__(
        self,
        cluster: ClusterData,
        object_id: int,
        mask: np.ndarray | None,
        recurse: int = 0,
    ):
        """Build from `ClusterData`, an object_id and mask specifying with sources
        in the cluster are part of the object"""
        self.parent_cluster: ClusterData = cluster
        self.object_id: int = object_id
        if mask is None:
            self.mask = np.ones((self.parent_cluster.n_src), dtype=bool)
        else:
            self.mask = mask
        self.catalog_id: np.ndarray = self.parent_cluster.catalog_id[self.mask]
        self.n_src: int = self.catalog_id.size
        self.n_unique: int = np.unique(self.catalog_id).size
        self.data: pandas.DataFrame | None = None

        self.recurse: int = recurse
        self.x_cent: float = np.nan
        self.y_cent: float = np.nan
        self.rms_dist: float = np.nan
        self.snr_mean: float = np.nan
        self.snr_rms: float = np.nan
        self.dist_2: np.ndarray = np.array([])
        self.extract()

    @property
    def x_cluster(self) -> np.ndarray:
        """Return the x-position of the sources within the footprint"""
        assert self.data is not None
        return self.data.x_cluster

    @property
    def y_cluster(self) -> np.ndarray:
        """Return the y-position of the sources within the footprint"""
        assert self.data is not None
        return self.data.y_cluster

    @property
    def x_pix(self) -> np.ndarray:
        """Return the x-position of the sources within the cell"""
        assert self.data is not None
        return self.data.x_pix

    @property
    def y_pix(self) -> np.ndarray:
        """Return the y-position of the sources within the cell"""
        assert self.data is not None
        return self.data.y_pix

    def hasRefCatalog(self, ref_cat_id: int = 0) -> bool:
        """Is there a source from the reference catalog"""
        return ref_cat_id in self.catalog_id

    def sourceIds(self) -> np.ndarray:
        """Return the source ids for the sources in the object"""
        return self.parent_cluster.src_id[self.mask]

    def sourceIdxs(self) -> np.ndarray:
        """Return the source indices for the sources in the object"""
        return self.parent_cluster.src_idx[self.mask]

    def _updateCatIndices(self) -> None:
        self.catalog_id = self.parent_cluster.catalog_id[self.mask]
        self.n_src = self.catalog_id.size
        self.n_unique = np.unique(self.catalog_id).size

    def extract(self) -> None:
        """Extract data from parent cluster"""
        assert self.parent_cluster.data is not None
        self.data = self.parent_cluster.data.iloc[self.mask]
        self._updateCatIndices()


class ShearObjectData(ObjectData):
    """Subclass of ObjectData that can compute shear statisitics"""

    def shearStats(self) -> dict:
        """Return the shear statistics"""
        assert self.data is not None
        return shear_utils.shearStats(self.data)
